"""Base class for ktables"""
import typing as t

import numpy as np
import numpy.typing as npt

if t.TYPE_CHECKING:
    from ..opacity import Opacity

    _Base = Opacity
else:
    _Base = object


class KTable(_Base):
    """A mixin class to represent a ktable."""

    @property
    def weights(self) -> npt.NDArray[np.float64]:
        """Quadrature weights."""
        raise NotImplementedError

    def opacity(
        self,
        temperature: float,
        pressure: float,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> npt.NDArray[np.float64]:
        """Opacity at given temperature and pressure.

        Parameters
        ----------
        temperature:
            Temperature in K
        pressure:
            Pressure in Pa
        wngrid:
            Wavenumber grid to interpolate to

        Returns
        -------
        np.ndarray:
            Interpolated opacity

        """
        from scipy.interpolate import interp1d

        wngrid_filter = slice(None)
        if wngrid is not None:
            wngrid_filter = np.where(
                (self.wavenumberGrid >= wngrid.min())
                & (self.wavenumberGrid <= wngrid.max())
            )[0]
        orig = self.compute_opacity(temperature, pressure, wngrid_filter).reshape(
            -1, len(self.weights)
        )

        if wngrid is None or np.array_equal(
            self.wavenumberGrid.take(wngrid_filter), wngrid
        ):
            return orig
        else:
            f = interp1d(
                self.wavenumberGrid[wngrid_filter],
                orig,
                axis=0,
                copy=False,
                bounds_error=False,
                fill_value=(orig[0], orig[-1]),
                assume_sorted=True,
            )
            return f(wngrid).reshape(-1, len(self.weights))





from .data.profiles.chemistry import ChemistryFile
from .data.profiles.chemistry.chemistry import Chemistry
from .data.profiles.chemistry import TaurexChemistry
from .data.profiles.chemistry.gas.gas import Gas
from .data.profiles.chemistry import ConstantGas
from .data.profiles.chemistry import TwoLayerGas
from .data.profiles.chemistry import PowerGas
from .data.profiles.chemistry.gas.arraygas import ArrayGas
from .data.profiles.chemistry import AutoChemistry



import typing as t

import numpy as np
import numpy.typing as npt

from taurex import OutputSize
from taurex.binning import BinDownType, BinnedSpectrumType, Binner
from taurex.types import ModelOutputType


class NativeBinner(Binner):
    """A `do-nothing` binner.

    This is useful when the pipeline expects
    a binner but none is given. Simplifies implementation and also
    handles dictionary writing of the forward model.
    """

    def bindown(
        self,
        wngrid: npt.NDArray[np.float64],
        spectrum: npt.NDArray[np.float64],
        grid_width: t.Optional[npt.NDArray[np.float64]] = None,
        error: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> BinDownType:
        """Bin down a spectrum from a high resolution to a lower resolution.

        This should handle the binning of a spectrum passed into the function.
        Parameters given are guidelines on expectation of usage.

        Parameters
        ----------
        wngrid : :obj:`array`
            The wavenumber grid of the spectrum to be binned down.
            Generally the 'native' wavenumber grid

        spectrum: :obj:`array`
            The spectra we wish to bin-down. Must be same shape as
            ``wngrid``.

        grid_width: :obj:`array`, optional
            Wavenumber grid full-widths for the spectrum to be binned down.
            Must be same shape as ``wngrid``.
            Optional, generally if you require this but the user does not pass
            it then you must compute it yourself using ``wngrid``. This can
            be done easily using the function
            func:`~taurex.util.compute_bin_edges`.

        error: :obj:`array`, optional
            Associated errors or noise of the spectrum. Must be same shape
            as ``wngrid``.Optional parameter, when implementing you must
            deal with the cases where either the error is passed or not passed.

        Returns
        -------
        binned_wngrid : :obj:`array`
            New wavenumber grid

        spectrum: :obj:`array`
            Binned spectrum.

        grid_width: :obj:`array`
            New grid-widths

        error: :obj:`array` or None
            If passed, should be the binned error otherwise None

        """
        return wngrid, spectrum, grid_width, error

    def bin_model(self, model_output: ModelOutputType) -> BinDownType:
        """Does nothing, only returns function arguments"""
        return *model_output[:2], None, None

    def generate_spectrum_output(
        self,
        model_output: ModelOutputType,
        output_size: t.Optional[OutputSize] = OutputSize.heavy,
    ) -> BinnedSpectrumType:
        """Generate spectrum output for lightcurves.

        Parameters
        ----------

        model_output: obj:`tuple`
            Result from running a TauREx3 forward model

        output_size: :class:`~taurex.taurexdefs.OutputSize`
            Size of the output.

        Returns
        -------
        :obj:`dict`:
            Dictionary of spectra containing both lightcurves
            and spectra.


        """
        output = {}

        wngrid, flux, tau, extra = model_output

        output["native_wngrid"] = wngrid
        output["native_wlgrid"] = 10000 / wngrid
        output["native_spectrum"] = flux

        if output_size > OutputSize.light:
            output["native_tau"] = tau

        return output



"""Main free chemistry model."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.core import FittingType, derivedparam
from taurex.exceptions import InvalidModelException
from taurex.output import OutputGroup
from taurex.util import has_duplicates, molecule_texlabel

from .autochemistry import AutoChemistry
from .gas.gas import Gas


class InvalidChemistryException(InvalidModelException):
    """Called when atmosphere mix is greater than unity."""

    pass


class TaurexChemistry(AutoChemistry):
    """The standard chemical model used in Taurex.

    This allows for the combinationof different mixing profiles for each molecule.
    Lets take an example
    profile, we want an atmosphere with a constant mixing of ``H2O`` but two
    layer mixing for ``CH4``.
    First we initialize our chemical model:

        >>> chemistry = TaurexChemistry()

    Then we can add our molecules using the :func:`addGas` method. Lets start
    with ``H2O``, since its a constant profile for all layers of the atmosphere
    we thus add
    the :class:`~taurex.data.profiles.chemistry.gas.constantgas.ConstantGas`
    object:

        >>> chemistry.addGas(ConstantGas('H2O',mix_ratio = 1e-4))

    Easy right? Now the same goes for ``CH4``, we can add the molecule into
    the chemical model by using the correct profile (in this case
    :class:`~taurex.data.profiles.chemistry.gas.twolayergas.TwoLayerGas`):

        >>> chemistry.addGas(TwoLayerGas('CH4',mix_ratio_surface=1e-4,
                                         mix_ratio_top=1e-8))

    Chaining is also supported:

        >>> chemistry.addGas(gas1).addGas(gas2)

    Molecular profiles available are:
        * :class:`~taurex.data.profiles.chemistry.gas.constantgas.ConstantGas`
        * :class:`~taurex.data.profiles.chemistry.gas.twolayergas.TwoLayerGas`
        * :class:`~taurex.data.profiles.chemistry.gas.twolayergas.TwoPointGas`


    """

    def __init__(
        self,
        fill_gases: t.Optional[t.List[str]] = None,
        ratio: t.Union[float, t.List[float]] = 0.17567,
        derived_ratios: t.List[str] = None,
        base_metallicity: float = 0.013,
    ) -> None:
        """Initialize free chemistry.

        Parameters
        ----------

        fill_gases : str or :obj:`list`
            Either a single gas or list of gases to fill the atmosphere with
            default: ``['H2','He']``

        ratio : float or :obj:`list`
            If a bunch of molecules are used to fill an atmosphere, whats the
            ratio between them?
            The first fill gas is considered the main one with others defined as
            ``molecule / main_molecule``
            default: ``0.17567``

        derived_ratios : :obj:`list`
            List of element ratios to compute as derived parameters
            (e.g. ``C/O``)

        base_metallicity : float
            What to consider the base metallicity of the atmosphere
            (default: 0.013)

        """
        super().__init__("ChemistryModel")

        self._gases: t.List[Gas] = []
        self._active = []
        self._inactive = []
        fill_gases = fill_gases or ["H2", "He"]
        derived_ratios = derived_ratios or []
        if isinstance(fill_gases, str):
            fill_gases = [fill_gases]

        if isinstance(ratio, float):
            ratio = [ratio]

        if has_duplicates(fill_gases):
            self.error("Fill gasses has duplicate molecules")
            self.error("Fill gasses: %s", fill_gases)
            raise ValueError("Duplicate fill gases detected")

        if len(fill_gases) > 1 and len(ratio) != len(fill_gases) - 1:
            self.error("Fill gases and ratio count are not correctly matched")
            self.error(
                "There should be %s ratios, you have defined %s",
                len(fill_gases) - 1,
                len(ratio),
            )
            raise InvalidChemistryException

        self._fill_gases = fill_gases
        self._fill_ratio = ratio
        self._mix_profile = None
        self._base_metallicity = base_metallicity
        self.debug("MOLECULES I HAVE %s", self.availableActive)
        self.setup_fill_params()
        self.determine_active_inactive()
        self.setup_derived_params(derived_ratios)

    # def determine_mix_mask(self):

    #     try:
    #         self._active, self._active_mask = zip(*[(m, i) for i, m in
    #                                                     enumerate(self.gases)
    #                                                     if m in
    #                                                     self.availableActive])
    #     except ValueError:
    #         self.debug('No active gases detected')
    #         self._active, self._active_mask = [], None

    #     try:
    #         self._inactive, self._inactive_mask = zip(*[(m, i) for i, m in
    #                                                         enumerate(self.gases)
    #                                                         if m not in
    #                                                         self.availableActive])
    #     except ValueError:
    #         self.debug('No inactive gases detected')
    #         self._inactive, self._inactive_mask = [], None

    #     self._active_mask = np.array(self._active_mask)
    #     self._inactive_mask = np.array(self._inactive_mask)

    def setup_fill_params(self) -> None:
        """Generate fill gas parameters."""
        if not hasattr(self._fill_gases, "__len__") or len(self._fill_gases) < 2:
            return

        main_gas = self._fill_gases[0]

        for idx, value in enumerate(zip(self._fill_gases[1:], self._fill_ratio)):
            gas, ratio = value
            mol_name = f"{gas}_{main_gas}"
            param_name = mol_name
            param_tex = "{}/{}".format(
                molecule_texlabel(gas), molecule_texlabel(main_gas)
            )

            def read_mol(self, idx=idx):
                return self._fill_ratio[idx]

            def write_mol(self, value, idx=idx):
                self._fill_ratio[idx] = value

            fget = read_mol
            fset = write_mol

            fget.__doc__ = f"{gas}/{main_gas} ratio (volume)"

            bounds = [1.0e-12, 0.1]

            default_fit = False
            self.add_fittable_param(
                param_name, param_tex, fget, fset, "log", default_fit, bounds
            )

    def setup_derived_params(self, ratio_list: t.List[str]) -> None:
        """Generate derived parameters."""
        for elem_ratio in ratio_list:
            elem1, elem2 = elem_ratio.split("/")
            mol_name = f"{elem1}_{elem2}_ratio"
            param_name = mol_name
            param_tex = "{}/{}".format(
                molecule_texlabel(elem1), molecule_texlabel(elem2)
            )

            def read_mol(self, elem=elem_ratio):
                return np.mean(self.get_element_ratio(elem))

            fget = read_mol

            fget.__doc__ = f"{elem_ratio} ratio (volume)"

            compute = True
            self.add_derived_param(param_name, param_tex, fget, compute)

    # def compute_mu_profile(self, nlayers):
    #     """
    #     Computes molecular weight of atmosphere
    #     for each layer

    #     Parameters
    #     ----------
    #     nlayers: int
    #         Number of layers
    #     """
    #     from taurex.util import get_molecular_weight
    #     self.mu_profile = np.zeros(shape=(nlayers,))
    #     if self.mixProfile is not None:
    #         mix_profile = self.mixProfile
    #         for idx, gasname in enumerate(self.gases):
    #             self.mu_profile += mix_profile[idx] * \
    #                 get_molecular_weight(gasname)

    def isActive(self, gas: str) -> bool:  # noqa: N802
        """Determines if the gas is absorbing.

        Parameters
        ----------

        gas: str
            Name of molecule


        Returns
        -------
        bool:
            True if active
        """
        if gas in self.availableActive:
            return True
        else:
            return False

    def addGas(self, gas: Gas) -> "TaurexChemistry":  # noqa: N802
        """Adds a gas in the atmosphere.

        Parameters
        ----------
        gas : :class:`~taurex.data.profiles.chemistry.gas.gas.Gas`
            Gas to add into the atmosphere. Only takes effect
            on next initialization call.

        """

        if gas.molecule in [x.molecule for x in self._gases]:
            self.error("Gas already exists %s", gas.molecule)
            raise ValueError("Gas already exists")

        self.debug("Gas %s fill gas: %s", gas.molecule, self._fill_gases)
        if gas.molecule in self._fill_gases:
            self.error(
                "Gas %s is already a fill gas: %s", gas.molecule, self._fill_gases
            )
            raise ValueError("Gas already exists")

        self._gases.append(gas)

        self.determine_active_inactive()

        return self

    @property
    def gases(self) -> t.List[str]:
        """List of gases in the atmosphere."""
        return self._fill_gases + [g.molecule for g in self._gases]

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mixing profile of all gases."""
        return self._mix_profile

    # @property
    # def activeGases(self):
    #     return self._active

    # @property
    # def inactiveGases(self):
    #     return self._inactive

    def compute_elements_mix(self) -> t.Dict[str, npt.NDArray[np.float64]]:
        """Determines the elemental mix of the atmosphere."""
        from taurex.util import split_molecule_elements

        element_dict = {}

        for g, m in zip(self.gases, self.mixProfile):
            avg_mix = m
            s = [], []
            if g != "e-":
                s = split_molecule_elements(g)
            else:
                s = {"e-": 1}
            # total_count = sum(s.values())

            for elements, count in s.items():
                val = element_dict.get(elements, 0.0)
                element_dict[elements] = val + count * avg_mix

        return element_dict

    @derivedparam(param_name="metallicity", param_latex="Z")
    def metallicity(self) -> float:
        """Metallicity of the atmosphere."""
        return self.get_metallicity()

    def get_metallicity(self) -> float:
        """Metallicity of the atmosphere.

        Determines metallicity of the atmosphere by computing the
        elemental mix of the atmosphere and then determining the
        metallicity by the following formula:

        .. math::

            Z = 1 - M_{H} - M_{He}

        where :math:`M_{H}` is the mass fraction of hydrogen and
        :math:`M_{He}` is the mass fraction of helium.

        Returns
        -------
        float:

        """
        from taurex.constants import AMU
        from taurex.util import mass

        element_dict = self.compute_elements_mix()

        total_mass = self.muProfile.sum() / AMU

        h_mass_fraction = element_dict["H"].sum() * mass["H"] / total_mass

        he_mass_fraction = element_dict["He"].sum() * mass["He"] / total_mass

        metallicity = 1 - h_mass_fraction - he_mass_fraction

        return metallicity / self._base_metallicity

    def get_element_ratio(self, elem_ratio: str) -> npt.NDArray[np.float64]:
        """Get element ratio of atmosphere.

        Parameters
        ----------
        elem_ratio : str
            Element ratio to compute of form ``elem1/elem2``

        Returns
        -------
        float:
            Element ratio
        """

        element_dict = self.compute_elements_mix()
        elem1, elem2 = elem_ratio.split("/")

        if elem1 not in element_dict:
            self.error(f"None of the gases have the element {elem1}")
            raise ValueError(f"No gas has element {elem1}")
        if elem2 not in element_dict:
            self.error(f"None of the gases have the element {elem2}")
            raise ValueError(f"No gas has element {elem2}")

        return element_dict[elem1].sum() / element_dict[elem2].sum()

    def fitting_parameters(self) -> t.Dict[str, FittingType]:
        """Overrides the fitting parameters to return
        one with all the gas profile parameters as well

        Returns
        -------

        fit_param:
            Dictionary of fitting parameters

        """
        full_dict = {}
        for gas in self._gases:
            full_dict.update(gas.fitting_parameters())

        full_dict.update(self._param_dict)

        return full_dict

    def initialize_chemistry(
        self,
        nlayers: int,
        temperature_profile: npt.NDArray[np.float64],
        pressure_profile: npt.NDArray[np.float64],
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize the chemistry.

        Parameters
        -----------
        nlayers : int
            number of layers
        temperature_profile : np.ndarray
            temperature profile
        pressure_profile : np.ndarray
            pressure profile
        altitude_profile : np.ndarray , optional
            altitude profile, deprecated

        """
        self.info("Initializing chemistry model")

        mix_profile = []

        for gas in self._gases:
            gas.initialize_profile(
                nlayers, temperature_profile, pressure_profile, altitude_profile
            )
            mix_profile.append(gas.mixProfile)

        total_mix = sum(mix_profile)

        self.debug("Total mix output %s", total_mix)

        validity = np.any(total_mix > 1.0)

        self.debug("Is invalid? %s", validity)

        if validity:
            self.error("Greater than 1.0 chemistry profile detected")
            raise InvalidChemistryException

        mixratio_remainder = 1.0 - total_mix

        mixratio_remainder += np.zeros(shape=(nlayers))
        mix_profile = self.fill_atmosphere(mixratio_remainder) + mix_profile

        if len(mix_profile) > 0:
            self._mix_profile = np.vstack(mix_profile)
        else:
            self._mix_profile = 0.0

        super().initialize_chemistry(
            nlayers, temperature_profile, pressure_profile, altitude_profile
        )

        super().compute_mu_profile(nlayers)

    def fill_atmosphere(self, mixratio_remainder: float) -> t.List[float]:
        """Fills the atmosphere with the fill gases."""
        fill = []

        if len(self._fill_gases) == 1:
            return [mixratio_remainder]
        else:
            main_molecule = mixratio_remainder * (1 / (1 + sum(self._fill_ratio)))

            fill.append(main_molecule)
            for _, ratio in zip(self._fill_gases[1:], self._fill_ratio):
                second_molecule = ratio * main_molecule
                fill.append(second_molecule)
        return fill

    # @property
    # def activeGasMixProfile(self):
    #     """
    #     Active gas layer by layer mix profile

    #     Returns
    #     -------
    #     active_mix_profile : :obj:`array`

    #     """
    #     return self.mixProfile[self._active_mask]

    # @property
    # def inactiveGasMixProfile(self):
    #     """
    #     Inactive gas layer by layer mix profile

    #     Returns
    #     -------
    #     inactive_mix_profile : :obj:`array`

    #     """
    #     return self.mixProfile[self._inactive_mask]

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write chemistry to output group."""
        gas_entry = super().write(output)
        if isinstance(self._fill_gases, float):
            gas_entry.write_scalar("ratio", self._fill_ratio)
        elif hasattr(self._fill_gases, "__len__"):
            gas_entry.write_array("ratio", np.array(self._fill_ratio))
        gas_entry.write_string_array("fill_gases", self._fill_gases)
        for gas in self._gases:
            gas.write(gas_entry)

        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for free chemistry."""
        return (
            "taurex",
            "free",
        )

    def citations(self):
        """Collect and return citations from all gases."""
        from taurex.data.citation import unique_citations_only

        old = super().citations()
        for g in self._gases:
            old.extend(g.citations())

        return unique_citations_only(old)



"""Output base class"""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.log import Logger
from taurex.types import AnyValType, ArrayType, ScalarType

MetadataType = t.Dict[str, AnyValType]
"""Type for metadata"""


class OutputGroup(Logger):
    """Stores output data in a hierarchical structure."""

    def __init__(self, name: str):
        super().__init__(name)
        self._name = name

    def write_array(
        self,
        array_name: str,
        array: ArrayType,
        metadata: t.Optional[MetadataType] = None,
    ) -> None:
        raise NotImplementedError

    def write_list(
        self,
        list_name: str,
        list_array: npt.ArrayLike,
        metadata: t.Optional[MetadataType] = None,
    ) -> None:
        arr = np.array(list_array)
        self.write_array(list_name, arr)

    def write_scalar(
        self,
        scalar_name: str,
        scalar: ScalarType,
        metadata: t.Optional[MetadataType] = None,
    ) -> None:
        raise NotImplementedError

    def write_string(
        self, string_name: str, string: str, metadata: t.Optional[MetadataType] = None
    ) -> None:
        raise NotImplementedError

    def write_string_array(
        self,
        string_name: str,
        string_array: t.Optional[str],
        metadata: t.Optional[MetadataType] = None,
    ) -> None:
        raise NotImplementedError

    def create_group(self, group_name: str) -> "OutputGroup":
        """Create a group."""
        raise NotImplementedError

    def store_dictionary(
        self, dictionary: t.Dict[str, t.Any], group_name: t.Optional[str] = None
    ) -> None:
        """Store a dictionary in the output."""
        from taurex.util import recursively_save_dict_contents_to_output

        group = self
        if group_name is not None:
            group = self.create_group(group_name)
        recursively_save_dict_contents_to_output(group, dictionary)


class Output(OutputGroup):
    """Base calss for handling outputs from Taurex3"""

    def __init__(self, name: str) -> None:
        """Initialize output.

        Parameters
        ----------
        name : str
            Name for logging purposes.

        """
        super().__init__(name)

    def open(self) -> None:
        """Open output."""
        raise NotImplementedError

    def close(self) -> None:
        """Close output."""
        raise NotImplementedError

    def __enter__(self) -> None:
        """Enter context manager."""
        self.open()
        return self

    def __exit__(self, type: t.Any, value: t.Any, traceback: t.Any) -> None:
        """Exit context manager."""
        self.close()

    def store_dictionary(
        self, dictionary: t.Dict[str, t.Any], group_name: t.Optional[str] = None
    ) -> None:
        """Store a dictionary in the output."""
        from taurex.util import recursively_save_dict_contents_to_output

        out = self
        if group_name is not None:
            out = self.create_group(group_name)

        recursively_save_dict_contents_to_output(out, dictionary)



"""Just contains a singleton class. Pretty useful"""


class Singleton:
    """A singleton for your usage.

    When inheriting do not implement __init__ instead override :func:`init`.
    """

    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        """Override to act as an init"""
        pass



"""Temperature profile loaded from array."""
import typing as t

import numpy as np
import numpy.typing as npt
from scipy.interpolate import interp1d

from taurex.output import OutputGroup

from .tprofile import TemperatureProfile


class TemperatureArray(TemperatureProfile):
    """Temperature profile loaded from array."""

    def __init__(
        self,
        tp_array: t.Optional[npt.ArrayLike] = None,
        p_points: t.Optional[npt.ArrayLike] = None,
        reverse: t.Optional[bool] = False,
    ):
        """Initialize the temperature profile.

        Parameters
        ----------
        tp_array:
            Temperature profile array, by default None
        p_points:
            Pressure profile array, by default None
        reverse : t.Optional[bool], optional
            Reverse the temperature profile, by default False

        """
        super().__init__(self.__class__.__name__)

        self._tp_profile = np.array(tp_array)
        if reverse:
            self._tp_profile = self._tp_profile[::-1]
        if p_points is not None:
            self._p_profile = np.array(p_points)
            if reverse:
                self._p_profile = self._p_profile[::-1]
            self._func = interp1d(
                np.log10(self._p_profile),
                self._tp_profile,
                bounds_error=False,
                fill_value=(self._tp_profile[-1], self._tp_profile[0]),
            )
        else:
            self._p_profile = None

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Returns temperature profile.

        Returns
        -------
        t_profile:
            temperature profile
        """
        if self._p_profile is None:
            if self._tp_profile.shape[0] == self.nlayers:
                return self._tp_profile
            interp_temp = np.linspace(1.0, 0.0, self._tp_profile.shape[0])
            interp_array = np.linspace(1.0, 0.0, self.nlayers)
            return np.interp(
                interp_array[::-1], interp_temp[::-1], self._tp_profile[::-1]
            )
        else:
            interp_array = np.log10(self.pressure_profile)
            return self._func(interp_array)

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write temperature profile to output."""
        temperature = super().write(output)

        temperature.write_scalar("tp_array", self._tp_profile)

        return temperature

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return all input keywords."""
        return (
            "array",
            "fromarray",
        )



# flake8: noqa
from taurex.core import Citable


class TauREXCitations(Citable):
    BIBTEX_ENTRIES = [
        r"""
@ARTICLE{2021ApJ...917...37A,
       author = {{Al-Refaie}, A.~F. and {Changeat}, Q. and {Waldmann}, I.~P. and {Tinetti}, G.},
        title = "{TauREx 3: A Fast, Dynamic, and Extendable Framework for Retrievals}",
      journal = {\apj},
     keywords = {Open source software, Astronomy software, Exoplanet atmospheres, Radiative transfer, Bayesian statistics, Planetary atmospheres, Planetary science, 1866, 1855, 487, 1335, 1900, 1244, 1255, Astrophysics - Instrumentation and Methods for Astrophysics, Astrophysics - Earth and Planetary Astrophysics},
         year = 2021,
        month = aug,
       volume = {917},
       number = {1},
          eid = {37},
        pages = {37},
          doi = {10.3847/1538-4357/ac0252},
archivePrefix = {arXiv},
       eprint = {1912.07759},
 primaryClass = {astro-ph.IM},
       adsurl = {https://ui.adsabs.harvard.edu/abs/2021ApJ...917...37A},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
        """,
        r"""
@ARTICLE{2022ApJ...932..123A,
       author = {{Al-Refaie}, A.~F. and {Changeat}, Q. and {Venot}, O. and {Waldmann}, I.~P. and {Tinetti}, G.},
        title = "{A Comparison of Chemical Models of Exoplanet Atmospheres Enabled by TauREx 3.1}",
      journal = {\apj},
     keywords = {Open source software, Publicly available software, Chemical abundances, Bayesian statistics, Exoplanet atmospheres, Exoplanet astronomy, Exoplanet atmospheric composition, Exoplanets, Radiative transfer, 1866, 1864, 224, 1900, 487, 486, 2021, 498, 1335, Astrophysics - Earth and Planetary Astrophysics, Astrophysics - Instrumentation and Methods for Astrophysics},
         year = 2022,
        month = jun,
       volume = {932},
       number = {2},
          eid = {123},
        pages = {123},
          doi = {10.3847/1538-4357/ac6dcd},
archivePrefix = {arXiv},
       eprint = {2110.01271},
 primaryClass = {astro-ph.EP},
       adsurl = {https://ui.adsabs.harvard.edu/abs/2022ApJ...932..123A},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    ]


taurex_citation = TauREXCitations()
__citations__ = taurex_citation.nice_citation()



"""Module for all forward models."""
from .directimage import DirectImageModel
from .emission import EmissionModel
from .model import ForwardModel
from .simplemodel import OneDForwardModel, SimpleForwardModel
from .transmission import TransmissionModel

__all__ = [
    "ForwardModel",
    "TransmissionModel",
    "SimpleForwardModel",
    "EmissionModel",
    "DirectImageModel",
    "OneDForwardModel",
]



"""Module for computing flat Mie opacity."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.model import OneDForwardModel
from taurex.output import OutputGroup

from .contribution import Contribution


class FlatMieContribution(Contribution):
    """Computes a flat (gray) absorption contribution.

    Absorption is computed as a flat value between two pressures
    across all wavenumbers.

    Parameters
    ----------

    flat_mix_ratio: float
        Opacity value

    flat_bottomP: float
        Bottom of absorbing region in Pa

    flat_topP: float
        Top of absorbing region in Pa

    """

    def __init__(
        self,
        flat_mix_ratio: float = 1e-10,
        flat_bottomP: float = -1,  # noqa: N803
        flat_topP: float = -1,
    ) -> None:
        super().__init__("Mie")

        self._mie_mix = flat_mix_ratio
        self._mie_bottom_pressure = flat_bottomP
        self._mie_top_pressure = flat_topP

    @fitparam(
        param_name="flat_topP",
        param_latex=r"$P^{mie}_\mathrm{top}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[1e-20, 1],
    )
    def mieTopPressure(self) -> float:  # noqa: N802
        """
        Pressure at top of absorbing region in Pa
        """
        return self._mie_top_pressure

    @mieTopPressure.setter
    def mieTopPressure(self, value: float) -> None:  # noqa: N802
        self._mie_top_pressure = value

    @fitparam(
        param_name="flat_bottomP",
        param_latex=r"$P^{mie}_\mathrm{bottom}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[1e-20, 1],
    )
    def mieBottomPressure(self) -> float:  # noqa: N802
        """Pressure at bottom of absorbing region in Pa."""
        return self._mie_bottom_pressure

    @mieBottomPressure.setter
    def mieBottomPressure(self, value: float) -> None:  # noqa: N802
        self._mie_bottom_pressure = value

    @fitparam(
        param_name="flat_mix_ratio",
        param_latex=r"$\chi_\mathrm{mie}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[1e-20, 1],
    )
    def mieMixing(self) -> float:  # noqa: N802
        """Opacity of absorbing region in :math:`m^2`."""
        return self._mie_mix

    @mieMixing.setter
    def mieMixing(self, value: float) -> None:  # noqa: N802
        self._mie_mix = value

    def prepare_each(
        self, model: OneDForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Computes and flat absorbing opacity for the pressure regions given.


        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``Flat`` and the weighted mie opacity.


        """
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]

        pressure_levels = np.log10(model.pressure.pressure_profile_levels[::-1])

        bottom_pressure = self.mieBottomPressure
        if bottom_pressure < 0:
            bottom_pressure = pressure_levels.max()

        top_pressure = np.log10(self.mieTopPressure)
        if top_pressure < 0:
            top_pressure = pressure_levels.min()

        p_left = pressure_levels[:-1]
        p_right = pressure_levels[1:]

        p_range = sorted([top_pressure, bottom_pressure])

        save_start = np.searchsorted(p_right, p_range[0], side="right")
        save_stop = np.searchsorted(p_left[1:], p_range[1], side="right")
        p_min = p_left[save_start : save_stop + 1]
        p_max = p_right[save_start : save_stop + 1]
        weight = np.minimum(p_range[-1], p_max) - np.maximum(p_range[0], p_min)
        weight /= weight.max()
        sigma_xsec = np.zeros(shape=(self._nlayers, wngrid.shape[0]))
        sigma_xsec[save_start : save_stop + 1] = weight[:, None] * self.mieMixing

        sigma_xsec = sigma_xsec[::-1]

        self.sigma_xsec = sigma_xsec

        yield "Flat", sigma_xsec

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write contribution to output.

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`
            Output object to write to

        Returns
        -------
        output: :class:`~taurex.output.output.Output`
            Output object that was written to

        """
        contrib = super().write(output)
        contrib.write_scalar("flat_mix_ratio", self._mie_mix)
        contrib.write_scalar("flat_bottomP", self._mie_bottom_pressure)
        contrib.write_scalar("flat_topP", self._mie_top_pressure)
        return contrib

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Return input keywords for the contribution."""
        return ("FlatMie",)



"""Two layer gas profile."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup
from taurex.util import molecule_texlabel, movingaverage

from .gas import Gas


class TwoLayerGas(Gas):
    """Two layer gas profile.

    A gas profile with two different mixing layers at the surface of the
    planet and top of the atmosphere seperated at a defined
    pressure point and smoothened.

    """

    def __init__(
        self,
        molecule_name: t.Optional[str] = "CH4",
        mix_ratio_surface: t.Optional[float] = 1e-4,
        mix_ratio_top: t.Optional[float] = 1e-8,
        mix_ratio_P: t.Optional[float] = 1e3,  # noqa: N803
        mix_ratio_smoothing: t.Optional[float] = 10,
    ):
        """Initialize a two layer gas profile.

        Parameters
        -----------
        molecule_name : str
        Name of molecule

        mix_ratio_surface : float
            Mixing ratio of the molecule on the planet surface

        mix_ratio_top : float
            Mixing ratio of the molecule at the top of the atmosphere

        mix_ratio_P : float
            Boundary Pressure point between the two layers

        mix_ratio_smoothing : float , optional
            smoothing window

        """
        super().__init__(self.__class__.__name__, molecule_name=molecule_name)

        if mix_ratio_smoothing <= 0:
            raise ValueError("Smoothing window must be positive")

        self._mix_surface = mix_ratio_surface or 1e-4
        self._mix_top = mix_ratio_top
        self._mix_ratio_pressure = mix_ratio_P
        self._mix_ratio_smoothing = mix_ratio_smoothing
        self._mix_profile = None
        self.add_surface_param()
        self.add_top_param()
        self.add_pressure_param()

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mixing ratio profile for molecule.

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """
        return self._mix_profile

    @property
    def mixRatioSurface(self) -> float:  # noqa: N802
        """Abundance on the planets surface."""
        return self._mix_surface

    @property
    def mixRatioTop(self) -> float:  # noqa: N802
        """Abundance on the top of atmosphere."""
        return self._mix_top

    @property
    def mixRatioPressure(self) -> float:  # noqa: N802
        """Pressure at which the abundance changes."""
        return self._mix_ratio_pressure

    @property
    def mixRatioSmoothing(self) -> float:  # noqa: N802
        """Smoothing window."""
        return self._mix_ratio_smoothing

    @mixRatioSurface.setter
    def mixRatioSurface(self, value: float) -> None:  # noqa: N802
        """Set abundance on the planets surface."""
        self._mix_surface = value

    @mixRatioTop.setter
    def mixRatioTop(self, value: float) -> None:  # noqa: N802
        """Set abundance on the top of atmosphere."""
        self._mix_top = value

    @mixRatioPressure.setter
    def mixRatioPressure(self, value: float) -> None:  # noqa: N802
        """Set pressure at which the abundance changes."""
        self._mix_pressure = value

    @mixRatioSmoothing.setter
    def mixRatioSmoothing(self, value: float) -> None:  # noqa: N802
        """Set smoothing window."""
        self._mix_smoothing = value

    def add_surface_param(self) -> None:
        """Generates surface fitting parameters.

        Has the form ``[Molecule]_surface``
        """
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_surface = f"{param_name}_surface"
        param_surf_tex = f"{param_tex}_surface"

        def read_surf(self):
            return self._mix_surface

        def write_surf(self, value):
            self._mix_surface = value

        read_surf.__doc__ = (
            f"Mixing ratio at the surface of the planet for {param_name}"
        )

        fget_surf = read_surf
        fset_surf = write_surf

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_surface,
            param_surf_tex,
            fget_surf,
            fset_surf,
            "log",
            default_fit,
            bounds,
        )

    def add_top_param(self) -> None:
        """Generates TOA fitting parameters.

        Has the form: ``[Molecule]_top``
        """
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_top = f"{param_name}_top"
        param_top_tex = f"{param_tex}_top"

        def read_top(self):
            return self._mix_top

        def write_top(self, value):
            self._mix_top = value

        read_top.__doc__ = f"Mixing ratio at the top of the atmosphere for {param_name}"

        fget_top = read_top
        fset_top = write_top

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_top, param_top_tex, fget_top, fset_top, "log", default_fit, bounds
        )

    def add_pressure_param(self) -> None:
        """Generates pressure fitting parameter.

        Has the form ``[Molecule]_P``
        """
        mol_name = self.molecule
        mol_tex = molecule_texlabel(mol_name)

        param_p = f"{mol_name}_P"
        param_p_tex = f"{mol_tex}_P"

        def read_p(self: "TwoLayerGas"):
            return self._mix_ratio_pressure

        def write_p(self: "TwoLayerGas", value: float):
            self._mix_ratio_pressure = value

        read_p.__doc__ = f"Pressure at which the mixing ratio changes for {mol_name}"

        fget_p = read_p
        fset_p = write_p

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_p, param_p_tex, fget_p, fset_p, "log", default_fit, bounds
        )

    def initialize_profile(
        self,
        nlayers: t.Optional[int] = None,
        temperature_profile: t.Optional[npt.NDArray[np.float64]] = None,
        pressure_profile: t.Optional[npt.NDArray[np.float64]] = None,
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ):
        """Initialize the profile.

        Parameters
        -----------
        nlayers : int
            number of layers
        temperature_profile : np.ndarray
            temperature profile
        pressure_profile : np.ndarray
            pressure profile
        altitude_profile : np.ndarray , optional
            altitude profile, deprecated

        """
        self._mix_profile = np.zeros(nlayers)

        smooth_window = self._mix_ratio_smoothing
        p_layer = np.abs(pressure_profile - self._mix_ratio_pressure).argmin()

        start_layer = max(int(p_layer - smooth_window / 2), 0)

        end_layer = min(int(p_layer + smooth_window / 2), nlayers - 1)

        p_nodes = [
            pressure_profile[0],
            pressure_profile[start_layer],
            pressure_profile[end_layer],
            pressure_profile[-1],
        ]

        c_nodes = [
            self.mixRatioSurface,
            self.mixRatioSurface,
            self.mixRatioTop,
            self.mixRatioTop,
        ]

        chemprofile = 10 ** np.interp(
            (np.log(pressure_profile[::-1])),
            np.log(p_nodes[::-1]),
            np.log10(c_nodes[::-1]),
        )

        wsize = nlayers * (smooth_window / 100.0)

        if wsize % 2 == 0:
            wsize += 1

        c_smooth = 10 ** movingaverage(np.log10(chemprofile), int(wsize))

        border = np.int((len(chemprofile) - len(c_smooth)) / 2)

        self._mix_profile = chemprofile[::-1]

        self._mix_profile[border:-border] = c_smooth[::-1]

    def write(self, output: OutputGroup):
        """Write gas profile to output."""
        gas_entry = super().write(output)
        gas_entry.write_scalar("mix_ratio_top", self.mixRatioTop)
        gas_entry.write_scalar("mix_ratio_surface", self.mixRatioSurface)
        gas_entry.write_scalar("mix_ratio_P", self.mixRatioPressure)
        gas_entry.write_scalar("mix_ratio_smoothing", self.mixRatioSmoothing)

        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, str]:
        return (
            "twolayer",
            "2layer",
        )

    BIBTEX_ENTRIES = [
        """
        @misc{changeat2019complex,
            title={Towards a more complex description of chemical profiles
            in exoplanets retrievals: A 2-layer parameterisation},
            author={Quentin Changeat and Billy Edwards and
                Ingo Waldmann and Giovanna Tinetti},
            year={2019},
            eprint={1903.11180},
            archivePrefix={arXiv},
            primaryClass={astro-ph.EP}
        }
        """,
    ]



"""Opacities loaded from pickle."""
import pathlib
import pickle  # noqa
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.mpi import allocate_as_shared
from taurex.types import PathLike
from taurex.util import sanitize_molecule_string

from .interpolateopacity import InterpModeType, InterpolatingOpacity


class PickleOpacity(InterpolatingOpacity):
    """Class for loading opacities from pickle files.

    This was the original format for TauREx 1/2 opacities. It is now
    deprecated in favour of HDF5 files.

    """

    @classmethod
    def discover(cls) -> t.List[t.Tuple[str, t.Tuple[pathlib.Path, str]]]:
        """Discover opacities from pickle files in path."""
        from taurex.cache import GlobalCache
        from taurex.util import sanitize_molecule_string

        path = GlobalCache()["xsec_path"]
        if path is None:
            return []
        path = pathlib.Path(path)

        files = path.glob("*.pickle")

        discovery = []

        interp = GlobalCache()["xsec_interpolation"] or "linear"

        for f in files:
            splits = f.stem.split(".")
            mol_name = sanitize_molecule_string(splits[0])

            discovery.append((mol_name, [f, interp]))

        return discovery

    def __init__(
        self,
        filename: PathLike,
        interpolation_mode: t.Optional[InterpModeType] = "linear",
    ) -> None:
        """Initialize and load pickle file.

        Parameters
        ----------
        filename: PathLike
            Path to pickle file

        interpolation_mode:
            Interpolation mode, by default "linear"

        Raises
        ------
        FileNotFoundError
            If file is not found

        """
        filename = pathlib.Path(filename)
        super().__init__(
            f"PickleOpacity:{filename.stem[0:10]}",
            interpolation_mode=interpolation_mode,
        )
        import warnings

        warnings.warn(
            "PickleOpacity is deprecated and will be removed"
            " in a future major version. "
            "Please use HDF5Opacity instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if not filename.exists():
            raise FileNotFoundError(f"File {filename} does not exist")
        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self._resolution = None
        self._load_pickle_file(filename)

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Molecule name."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Opacity grid."""
        return self._xsec_grid

    def _load_pickle_file(self, filename: pathlib.Path) -> None:
        """Load pickle file into memory."""
        # Load the pickle file
        self.info(f"Loading opacity from {filename}")
        try:
            with open(filename, "rb") as f:
                self._spec_dict = pickle.load(f)  # noqa
        except UnicodeDecodeError:
            with open(filename, "rb") as f:
                self._spec_dict = pickle.load(f, encoding="latin1")  # noqa

        self._wavenumber_grid = self._spec_dict["wno"]

        self._temperature_grid = self._spec_dict["t"]
        self._pressure_grid = self._spec_dict["p"] * 1e5
        self._xsec_grid = allocate_as_shared(self._spec_dict["xsecarr"], logger=self)
        self._resolution = np.average(np.diff(self._wavenumber_grid))

        splits = filename.stem.split(".")
        mol_name = sanitize_molecule_string(splits[0])
        self._molecule_name = mol_name

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()
        self.clean_molecule_name()

    def clean_molecule_name(self) -> None:
        """Clean molecule name from underscores used."""
        splits = self.moleculeName.split("_")
        self._molecule_name = splits[0]

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        return self._pressure_grid

    @property
    def resolution(self) -> float:  # noqa: N802
        """Resolution of opacity."""
        return self._resolution



"""Gas from an array."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup

from .gas import Gas


class ArrayGas(Gas):
    """Gas profile from an array.

    Molecular abundance is interpolated if the number of layers do not match

    """

    def __init__(
        self,
        molecule_name: t.Optional[str] = "H2O",
        mix_ratio_array: t.Optional[t.List[float]] = None,
    ) -> None:
        super().__init__(self.__class__.__name__, molecule_name)
        """Initialize gas profile from an array.

        Parameters
        -----------
        molecule_name : str
            Name of molecule

        mix_ratio_array : :obj:`array`
            Mixing ratio of the molecule at each layer


        """
        mix_ratio_array = mix_ratio_array or [1e-2, 1e-6]
        self._mix_ratio_array = np.array(mix_ratio_array)

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mixing profile.

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """

        return self._mix_array

    def initialize_profile(
        self,
        nlayers: t.Optional[int] = None,
        temperature_profile: t.Optional[npt.NDArray[np.float64]] = None,
        pressure_profile: t.Optional[npt.NDArray[np.float64]] = None,
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize the gas profile.

        Parameters
        -----------
        nlayers : int
            Number of layers in the atmosphere
        temperature_profile : :obj:`array`
            Temperature profile of the atmosphere
        pressure_profile : :obj:`array`
            Pressure profile of the atmosphere
        altitude_profile : :obj:`array`, optional
            Altitude profile of the atmosphere (Deprecated)

        """
        if nlayers is None:
            self.error("number layers argument required")
            raise ValueError("Number of layers argument required")

        interp_array = np.linspace(0.0, 1.0, self._mix_ratio_array.shape[0])

        layer_interp = np.linspace(0.0, 1.0, nlayers)

        self._mix_array = np.interp(layer_interp, interp_array, self._mix_ratio_array)

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write parameters to output.


        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`
        """
        gas_entry = super().write(output)
        gas_entry.write_array("mix_ratio_array", self._mix_ratio_array)

        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Input keywords for this class."""
        return (
            "array",
            "fromarray",
        )



"""General utility functions."""
import re
import typing as t

import numpy as np
import numpy.typing as npt
from astropy import units as u

from taurex.output.output import OutputGroup
from taurex.types import AnyValType, ScalarType

mass = {
    "H": 1.00794,
    "He": 4.002602,
    "Li": 6.941,
    "Be": 9.012182,
    "B": 10.811,
    "C": 12.011,
    "N": 14.00674,
    "O": 15.9994,
    "F": 18.9984032,
    "Ne": 20.1797,
    "Na": 22.989768,
    "Mg": 24.3050,
    "Al": 26.981539,
    "Si": 28.0855,
    "P": 30.973762,
    "S": 32.066,
    "Cl": 35.4527,
    "Ar": 39.948,
    "K": 39.0983,
    "Ca": 40.078,
    "Sc": 44.955910,
    "Ti": 47.88,
    "V": 50.9415,
    "Cr": 51.9961,
    "Mn": 54.93805,
    "Fe": 55.847,
    "Co": 58.93320,
    "Ni": 58.6934,
    "Cu": 63.546,
    "Zn": 65.39,
    "Ga": 69.723,
    "Ge": 72.61,
    "As": 74.92159,
    "Se": 78.96,
    "Br": 79.904,
    "Kr": 83.80,
    "Rb": 85.4678,
    "Sr": 87.62,
    "Y": 88.90585,
    "Zr": 91.224,
    "Nb": 92.90638,
    "Mo": 95.94,
    "Tc": 98,
    "Ru": 101.07,
    "Rh": 102.90550,
    "Pd": 106.42,
    "Ag": 107.8682,
    "Cd": 112.411,
    "In": 114.82,
    "Sn": 118.710,
    "Sb": 121.757,
    "Te": 127.60,
    "I": 126.90447,
    "Xe": 131.29,
    "Cs": 132.90543,
    "Ba": 137.327,
    "La": 138.9055,
    "Ce": 140.115,
    "Pr": 140.90765,
    "Nd": 144.24,
    "Pm": 145,
    "Sm": 150.36,
    "Eu": 151.965,
    "Gd": 157.25,
    "Tb": 158.92534,
    "Dy": 162.50,
    "Ho": 164.93032,
    "Er": 167.26,
    "Tm": 168.93421,
    "Yb": 173.04,
    "Lu": 174.967,
    "Hf": 178.49,
    "Ta": 180.9479,
    "W": 183.85,
    "Re": 186.207,
    "Os": 190.2,
    "Ir": 192.22,
    "Pt": 195.08,
    "Au": 196.96654,
    "Hg": 200.59,
    "Tl": 204.3833,
    "Pb": 207.2,
    "Bi": 208.98037,
    "Po": 209,
    "At": 210,
    "Rn": 222,
    "Fr": 223,
    "Ra": 226.0254,
    "Ac": 227,
    "Th": 232.0381,
    "Pa": 213.0359,
    "U": 238.0289,
    "Np": 237.0482,
    "Pu": 244,
    "Am": 243,
    "Cm": 247,
    "Bk": 247,
    "Cf": 251,
    "Es": 252,
    "Fm": 257,
    "Md": 258,
    "No": 259,
    "Lr": 260,
    "Rf": 261,
    "Db": 262,
    "Sg": 263,
    "Bh": 262,
    "Hs": 265,
    "Mt": 266,
    "e-": 5.4857990907e-4,
}

ElementType = t.Dict[str, int]


def calculate_weight(chem: str) -> float:
    """Compute the molecular weight of a molecule in amu.

    Parameters
    ----------
    chem : str
        Molecule name e.g. H2O, CO2, CH4, etc.

    Returns
    -------
    float
        Molecular weight in amu

    """
    s = split_molecule_elements(chem)
    compoundweight = 0.0
    for element, count in s.items():
        compoundweight += mass[element] * count
    return compoundweight


# def split_molecule_elements(chem):
#     s = re.findall('([A-Z][a-z]?)([0-9]*)', chem)
#     return s


def tokenize_molecule(molecule: str) -> list[str]:
    """Tokenize a molecule string into its elements and numbers."""
    import re

    return re.findall(r"[A-Z][a-z]?|\d+|.", molecule)


def merge_elements(
    elem1: ElementType, elem2: ElementType, factor: t.Optional[int] = 1
) -> ElementType:
    """Merge two element dictionaries."""
    return {
        elem: elem1.get(elem, 0) + elem2.get(elem, 0) * factor
        for elem in set(elem1) | set(elem2)
    }


def split_molecule_elements(  # noqa: C901
    molecule: t.Optional[str] = None, tokens: t.Optional[t.List[str]] = None
) -> ElementType:
    """Split a molecule string into its elements and numbers.

    For example when run with H2O:

    >>> split_molecule_elements('H2O')
    {'H': 2, 'O': 1}

    Parameters
    ----------
    molecule : str, optional
        Molecule string to split

    tokens : list[str], optional
        List of presplit tokens.

    Returns
    -------
    dict[str, int]
        Dictionary of elements and their counts

    """
    from taurex.util import mass

    elems = {}

    if molecule:
        tokens = tokenize_molecule(molecule)

    length = 0

    while length < len(tokens):
        token = tokens[length]

        if token in mass:
            if token not in elems:
                elems[token] = 0
            try:
                peek = int(tokens[length + 1])

                length += 1
            except IndexError:
                peek = 1
            except ValueError:
                peek = 1
            elems[token] += peek
        elif token in "{([":  # noqa: S105
            length += 1
            sub_elems, moved = split_molecule_elements(tokens=tokens[length:])
            length += moved
            try:
                peek = int(tokens[length + 1])
                length += 1
            except IndexError:
                peek = 1
            except ValueError:
                peek = 1
            elems = merge_elements(elems, sub_elems, peek)
        elif token in "}])":  # noqa: S105
            return elems, length
        length += 1

    return elems


def sanitize_molecule_string(molecule: str) -> str:
    """Cleans a molecule string to match up with molecule naming in TauREx3.

    For example:

    >>> sanitize_molecule_string('H2O')
    'H2O'
    >>> sanitize_molecule_string('H2-16O')
    'H2O'


    Parameters
    ----------
    molecule: str
        Molecule to sanitize

    Returns
    -------
    str:
        Sanitized name

    """
    return "".join(["".join(s) for s in re.findall("([A-Z][a-z]?)([0-9]*)", molecule)])


_mol_latex = {
    "HE": "He",
    "H2": "H$_2$",
    "N2": "N$_2$",
    "O2": "O$_2$",
    "CO2": "CO$_2$",
    "CH4": "CH$_4$",
    "CO": "CO",
    "NH3": "NH$_3$",
    "H2O": "H$_2$O",
    "C2H2": "C$_2$H$_2$",
    "HCN": "HCN",
    "H2S": "H$_2$S",
    "SIO2": "SiO$_2$",
    "SO2": "SO$_2$",
}
"""Latex versions of molecule names"""


def get_molecular_weight(gasname: str) -> float:
    """For a given molecule return the molecular weight in kg

    Parameters
    ----------
    gasname : str
        Name of molecule

    Returns
    -------
    float :
        molecular weight in amu or 0 if not found

    """
    from taurex.constants import AMU

    mu = calculate_weight(gasname)

    return mu * AMU


# TODO: Generalize this to any molecule.
def molecule_texlabel(gasname: str) -> str:
    """For a given molecule return its latex form


    Parameters
    ----------
    gasname : str
        Name of molecule

    Returns
    -------
    str :
        Latex form of the molecule or just the passed name if not found


    """
    gasname = gasname

    try:
        return _mol_latex[gasname]
    except KeyError:
        return gasname


def bindown(
    original_bin: npt.NDArray, original_data: npt.NDArray, new_bin: npt.NDArray
) -> npt.NDArray:
    """This method quickly bins down by taking the mean.

    The numpy histogram function is exploited to do this quickly.
    This is prone to nans if no data is present in the bin.

    Parameters
    ----------
    original_bin: :obj:`numpy.array`
        The original bins for the that we want to bin down

    original_data: :obj:`numpy.array`
        The associated data that will be averaged along the new bins

    new_bin: :obj:`numpy.array`
        The new binnings we want to use (must have less points than the original)

    Returns
    -------
    :obj:`array`
        Binned mean of ``original_data``


    """
    import numpy as np

    # print(original_bin.shape,original_data.shape)
    # if last_point is None:
    #    last_point = new_bin[-1]*2
    # calc_bin = np.append(new_bin,last_point)
    # return(np.histogram(original_bin, calc_bin, weights=original_data)[0] /
    #          np.histogram(original_bin,calc_bin)[0])

    filter_lhs = np.zeros(new_bin.shape[0] + 1)
    filter_lhs[0] = new_bin[0]
    filter_lhs[0] -= (new_bin[1] - new_bin[0]) / 2
    filter_lhs[-1] = new_bin[-1]
    filter_lhs[-1] += (new_bin[-1] - new_bin[-2]) / 2
    filter_lhs[1:-1] = (new_bin[1:] + new_bin[:-1]) / 2
    axis = len(original_data.shape) - 1
    if axis:
        digitized = np.digitize(original_bin, filter_lhs, right=True)
        axis = len(original_data.shape) - 1
        bin_means = [
            original_data[..., digitized == i].mean(axis=axis)
            for i in range(1, len(filter_lhs))
        ]
        return np.column_stack(bin_means)
    return (
        np.histogram(original_bin, filter_lhs, weights=original_data)[0]
        / np.histogram(original_bin, filter_lhs)[0]
    )


def movingaverage(a: npt.NDArray, n: t.Optional[int] = 3) -> npt.NDArray:
    """Computes moving average given an array and window size.

    Parameters
    ----------
    a : :obj:`array`
        Array to compute average

    n : int
        Averaging window

    Returns
    -------
    :obj:`array`
        Resultant array

    """
    import numpy as np

    ret = np.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1 :] / n


def quantile_corner(
    x: npt.NDArray,
    q: t.Union[npt.NDArray, float],
    weights: t.Optional[t.Union[float, npt.NDArray]] = None,
) -> npt.NDArray:
    """Compute quantiles from an array with weighting.

    * Taken from corner.py
    __author__ = "Dan Foreman-Mackey (danfm@nyu.edu)"
    __copyright__ = "Copyright 2013-2015 Daniel Foreman-Mackey"

    Like numpy.percentile, but:

    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x

    Parameters
    ----------

    x : :obj:`array`
        Input array or object that can be converted to an array.

    q : :obj:`array` or float
        Percentile or sequence of percentiles to compute, which
        must be between 0 and 1 inclusive.

    weights : :obj:`array` or float , optional
        Weights on x

    Returns
    -------
    percentile : scalar or ndarray


    """
    import numpy as np

    if weights is None:
        return np.percentile(x, [100.0 * qi for qi in q])
    else:
        idx = np.argsort(x)
        xsorted = x[idx]
        cdf = np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return np.interp(q, cdf, xsorted).tolist()


def loadtxt2d(intext: str) -> npt.NDArray:
    """Wraps loadtext.

    Returns a 2d array or 1d array depending on the input text.

    Parameters
    ----------
    intext : str
        Input text

    Returns
    -------
    :obj:`array`
        2d array or 1d array

    """
    try:
        return np.loadtxt(intext, ndmin=2)
    except Exception:
        return np.loadtxt(intext)


def read_error_line(line: str) -> t.Tuple[str, float, float]:
    """Reads line from multinest"""
    print("_read_error_line -> line>", line)
    name, values = line.split("   ", 1)
    print("_read_error_line -> name>", name)
    print("_read_error_line -> values>", values)
    name = name.strip(": ").strip()
    values = values.strip(": ").strip()
    v, error = values.split(" +/- ")
    return name, float(v), float(error)


def read_error_into_dict(line: str, d: t.Dict[str, float]) -> None:
    """Reads multinest error into dict."""
    name, v, error = read_error_line(line)
    d[name.lower()] = v
    d["%s error" % name.lower()] = error


def read_table(
    txt: str, d: t.Optional[t.Dict[str, float]] = None, title: t.Optional[str] = None
):
    """Reads a table into a dictionary from multinest outputs."""
    from io import StringIO

    import numpy as np

    if title is None:
        title, table = txt.split("\n", 1)
    else:
        table = txt
    header, table = table.split("\n", 1)
    data = loadtxt2d(StringIO(table))
    if d is not None:
        d[title.strip().lower()] = data
    if len(data.shape) == 1:
        data = np.reshape(data, (1, -1))
    return data


def decode_string_array(f):
    """Helper to decode strings from hdf5."""
    sl = list(f)
    return [s[0].decode("utf-8") for s in sl]


OutputItem = t.Union[
    float, int, np.int64, np.float64, np.ndarray, str, t.List, t.Tuple, dict
]


def recursively_save_dict_contents_to_output(
    output: OutputGroup, dic: t.Dict[str, OutputItem]
):
    """Will recursive write a dictionary into output.

    Parameters
    ----------
    output:
        Group (or root) in output file to write to

    dic: dict
        Dictionary we want to write

    Raises
    ------
    ValueError
        If item is not a supported type

    """
    for key, item in dic.items():
        try:
            store_thing(output, key, item)
        except TypeError as e:
            raise ValueError("Cannot save %s type" % type(item)) from e


def store_thing(output: OutputGroup, key: str, item: OutputItem) -> None:  # noqa: C901
    """Stores a single item into output.

    Parameters
    ----------
    output:
        Group (or root) in output file to write to
    key: str
        Name of item
    item:
        Item to store

    Raises
    ------
    TypeError
        If item is not a supported type

    """
    if isinstance(
        item,
        (
            float,
            int,
            np.int64,
            np.float64,
        ),
    ):
        output.write_scalar(key, item)
    elif isinstance(item, (np.ndarray,)):
        output.write_array(key, item)
    elif isinstance(item, (str,)):
        output.write_string(key, item)
    elif isinstance(
        item,
        (
            list,
            tuple,
        ),
    ):
        if isinstance(item, tuple):
            item = list(item)
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, item)
        else:
            try:
                output.write_array(key, np.array(item))

            except TypeError:
                for idx, val in enumerate(item):
                    new_key = f"{key}{idx}"
                    store_thing(output, new_key, val)

    elif isinstance(item, dict):
        group = output.create_group(key)
        recursively_save_dict_contents_to_output(group, item)
    else:
        raise TypeError


def weighted_avg_and_std(
    values: npt.ArrayLike, weights: npt.ArrayLike, axis: t.Optional[int] = None
) -> t.Tuple[AnyValType, AnyValType]:
    """Computes weight average and standard deviation.

    Parameters
    ----------
    values : :obj:`array`
        Input array

    weights : :obj:`array`
        Must be same shape as ``values``


    axis : int , optional
        axis to perform weighting

    """
    import numpy as np

    average = np.average(values, weights=weights, axis=axis)
    variance = np.average(
        (values - average) ** 2, weights=weights, axis=axis
    )  # Fast and numerically precise
    return (average, np.sqrt(variance))


def random_int_iter(total: int, fraction: t.Optional[float] = 1.0) -> t.Iterator[int]:
    """Iterator to randomly sample integers up to a total number.

    Fraction is the fraction of total to sample. For example
    if total = 100 and fraction = 0.1 then 10 random integers
    will be sampled between 0 and 99.



    Parameters
    ----------
    total : int
        Maximum number
    fraction : float
        Fraction of total to sample

    Yields
    ------
    int
        Random integer

    """
    import random

    n_points = int(total * fraction)

    samples = random.sample(range(total), n_points)
    yield from samples


def compute_bin_edges(spectral_grid: npt.NDArray) -> t.Tuple[npt.NDArray, npt.NDArray]:
    """Computes bin edges from a spectral grid.

    Parameters
    ----------
    spectral_grid : :obj:`array`
        Spectral grid

    Returns
    -------
    :obj:`array`
        Bin edges
    :obj:`array`
        Bin widths

    """
    import numpy as np

    diff = np.diff(spectral_grid) / 2
    edges = np.concatenate(
        [
            [spectral_grid[0] - (spectral_grid[1] - spectral_grid[0]) / 2],
            spectral_grid[:-1] + diff,
            [(spectral_grid[-1] - spectral_grid[-2]) / 2 + spectral_grid[-1]],
        ]
    )
    return edges, np.abs(np.diff(edges))


def clip_native_to_wngrid(
    native_grid: npt.NDArray, spectral: npt.NDArray
) -> npt.NDArray:
    """Clips native grid values to a different spectral grid.

    Parameters
    ----------
    native_grid : :obj:`array`
        Native spectral grid

    spectral : :obj:`array`
        spectral grid

    Returns
    -------
    :obj:`array`
        Clipped native spectral grid

    """
    min_spectral = spectral.min()
    max_spectral = spectral.max()
    # Compute the maximum width
    wnwidths = compute_bin_edges(spectral)[-1]
    wn_min = min_spectral - wnwidths.max()
    wn_max = max_spectral + wnwidths.max()

    native_filter = (native_grid >= wn_min) & (native_grid <= wn_max)
    return native_grid[native_filter]


def wnwidth_to_wlwidth(wngrid: npt.NDArray, wnwidth: npt.NDArray) -> npt.NDArray:
    """Converts a wavenumber width to wavelength width and vice versa.

    Given a spectral grid and its associated spectral bin widths, this
    function will convert the wavenumber widths to wavelength widths and
    vice versa.

    The formula used is:

    .. math::
        \\Delta \\lambda = \\frac{10000 \\Delta \\nu}{\\nu^2}



    Parameters
    ----------
    wngrid : :obj:`array`
        Wavenumber grid in :math:`cm^{-1}`

    wnwidth : :obj:`array`
        Wavenumber width in :math:`cm^{-1}`

    Returns
    -------
    :obj:`array`
        Wavelength width in :math:`\\mu m`

    """
    return 10000 * wnwidth / (wngrid**2)


def class_from_keyword(keyword, class_filter=None):
    from ..parameter.classfactory import ClassFactory

    cf = ClassFactory()

    combined_classes = []
    if class_filter is None:
        combined_classes = (
            list(cf.temperatureKlasses)
            + list(cf.pressureKlasses)
            + list(cf.chemistryKlasses)
            + list(cf.gasKlasses)
            + list(cf.planetKlasses)
            + list(cf.starKlasses)
            + list(cf.modelKlasses)
            + list(cf.contributionKlasses)
        )
    else:
        if hasattr(class_filter, "__len__"):
            for x in class_filter:
                combined_classes += list(cf.list_from_base(x))
        else:
            combined_classes = list(cf.list_from_base(class_filter))

    for x in combined_classes:
        try:
            if keyword in x.input_keywords():
                return x
        except NotImplementedError:
            continue

    return None


def class_for_name(class_name: str):
    """Converts a string to a class.

    Searches TauREx3 registry of classes (including plugins) for name.



    Parameters
    ----------
    class_name : str
        Name of class

    """
    from ..parameter.classfactory import ClassFactory

    cf = ClassFactory()

    combined_classes = (
        list(cf.temperatureKlasses)
        + list(cf.pressureKlasses)
        + list(cf.chemistryKlasses)
        + list(cf.gasKlasses)
        + list(cf.planetKlasses)
        + list(cf.starKlasses)
        + list(cf.modelKlasses)
        + list(cf.contributionKlasses)
    )

    try:
        class_name = class_name.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    combined_classes_name = [c.__name__ for c in combined_classes]

    if class_name in combined_classes_name:
        return combined_classes[combined_classes_name.index(class_name)]
    else:
        raise Exception(f"Class of name {class_name} does not exist")


def create_grid_res(
    resolution: ScalarType, spectral_min: ScalarType, spectral_max: ScalarType
) -> npt.NDArray:
    """Creates a grid with a given resolution.

    Resolution is defined as :math:`R = \\frac{\\lambda}{\\Delta \\lambda}`

    Parameters
    ----------
    resolution : float
        Resolution to use
    spectral_min : float
        Minimum wavelength
    spectral_max : float
        Maximum wavelength

    Returns
    -------
    :obj:`array`
        Grid with resolution and spectral bin widths

    """
    #
    # R = l/Dl
    # l = (l-1)+Dl/2 + (Dl-1)/2
    #
    # --> (R - 1/2)*Dl = (l-1) + (Dl-1)/2
    #
    #
    spectral_list = []
    width_list = []
    wave = spectral_min
    width = wave / resolution

    while wave < spectral_max:
        width = wave / (resolution - 0.5) + width / 2 / (resolution - 0.5)
        wave = resolution * width
        width_list.append(width)
        spectral_list.append(wave)

    return np.array((spectral_list, width_list)).T


def conversion_factor(from_unit: str, to_unit: str) -> float:
    """Determine conversion from one unit to another.

    Parameters
    ----------
    from_unit : :class:`~astropy.units.Unit`
        Unit to convert from

    to_unit : :class:`~astropy.units.Unit`
        Unit to convert to

    Returns
    -------
    float
        Conversion factor.

    """
    try:
        from_conv = u.Unit(from_unit)
    except ValueError:
        from_conv = u.Unit(from_unit, format="cds")

    try:
        to_conv = u.Unit(to_unit)
    except ValueError:
        to_conv = u.Unit(to_unit, format="cds")

    return from_conv.to(to_conv)


def compute_dz(altitude: npt.NDArray) -> npt.NDArray:
    dz = np.zeros_like(altitude)
    dz[:-1] = np.diff(altitude)
    dz[-1] = altitude[-1] - altitude[-2]

    return dz


def has_duplicates(arr: npt.ArrayLike) -> bool:
    """Checks if an array has duplicates."""
    return len(arr) != len(set(arr))


def find_closest_pair(arr, value) -> (int, int):
    """Will find the indices that lie to the left and right of the value.

    `arr[left] <= value <= arr[right]`

    If the value is less than the array minimum then it will
    always return left=0 and right=1

    If the value is above the maximum

    Parameters
    ----------
    arr: :obj:`array`
        Array to search, must be sorted

    value: float
        Value to find in array


    Returns
    -------
    left: int

    right: int

    """

    right = arr.searchsorted(value)
    right = max(min(arr.shape[0] - 1, right), 1)

    left = right - 1
    left = max(0, left)

    return left, right


def ensure_string_utf8(val: str) -> str:
    """Ensures a string is utf8 encoded."""
    output = val

    try:
        output = val.decode()
    except (
        UnicodeDecodeError,
        AttributeError,
    ):
        pass

    return output



"""Transit forward model."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.chemistry import Chemistry
from taurex.planet import Planet
from taurex.pressure import PressureProfile
from taurex.stellar import Star
from taurex.temperature import TemperatureProfile

from .simplemodel import OneDForwardModel

if t.TYPE_CHECKING:
    from taurex.contributions import Contribution
else:
    Contribution = object


class TransmissionModel(OneDForwardModel):
    """A forward model for transits."""

    def __init__(
        self,
        planet: t.Optional[Planet] = None,
        star: t.Optional[Star] = None,
        pressure_profile: t.Optional[PressureProfile] = None,
        temperature_profile: t.Optional[TemperatureProfile] = None,
        chemistry: t.Optional[Chemistry] = None,
        nlayers: t.Optional[int] = 100,
        atm_min_pressure: t.Optional[float] = 1e-4,
        atm_max_pressure: t.Optional[float] = 1e6,
        contributions: t.Optional[t.List[Contribution]] = None,
        new_path_method: t.Optional[bool] = False,
    ) -> None:
        """Initialize transit forward model.

        Parameters
        ----------
        name: str
            Name to use in logging

        planet:
            Planet model, default planet is Jupiter

        star:
            Star model, default star is Sun-like

        pressure_profile:
            Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
            and ``atm_max_pressure``

        temperature_profile:
            Temperature model, default is an
            :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
            profile at 1500 K

        chemistry:
            Chemistry model, default is
            :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry`
            with ``H2O`` and ``CH4``

        nlayers: int, optional
            Number of layers. Used if ``pressure_profile`` is not defined.

        atm_min_pressure: float, optional
            Pressure at TOA. Used if ``pressure_profile`` is not defined.

        atm_max_pressure: float, optional
            Pressure at BOA. Used if ``pressure_profile`` is not defined.

        contributions: list, optional
            List of contributions to include

        new_path_method: bool, optional
            Use new path length computation method

        """
        super().__init__(
            self.__class__.__name__,
            planet,
            star,
            pressure_profile,
            temperature_profile,
            chemistry,
            nlayers,
            atm_min_pressure,
            atm_max_pressure,
            contributions,
        )
        self.new_method = new_path_method

    def compute_path_length_old(
        self, dz: npt.NDArray[np.float64]
    ) -> t.List[npt.NDArray[np.float64]]:
        """Compute path length for each layer."""
        dl = []

        planet_radius = self._planet.fullRadius
        total_layers = self.nLayers

        z = self.altitudeProfile
        self.debug("Computing path_length: \n z=%s \n dz=%s", z, dz)

        for layer in range(0, total_layers):
            p = (planet_radius + dz[0] / 2 + z[layer]) ** 2
            k = np.zeros(shape=(self.nLayers - layer))
            k[0] = np.sqrt(
                (planet_radius + dz[0] / 2.0 + z[layer] + dz[layer] / 2.0) ** 2 - p
            )

            k[1:] = np.sqrt(
                (planet_radius + dz[0] / 2 + z[layer + 1 :] + dz[layer + 1 :] / 2) ** 2
                - p
            )

            k[1:] -= np.sqrt(
                (
                    planet_radius
                    + dz[0] / 2
                    + z[layer : self.nLayers - 1]
                    + dz[layer : self.nLayers - 1] / 2
                )
                ** 2
                - p
            )

            dl.append(k * 2.0)
        return dl

    def compute_path_length(self) -> t.List[npt.NDArray[np.float64]]:
        """Compute path length for each layer, new method."""
        from taurex.util.geometry import parallel_vector

        altitude_boundaries = self.altitude_boundaries
        radius = self.planet.fullRadius

        # Generate our line of sight paths
        viewer, tangent = parallel_vector(
            radius, self.altitude_profile + self.deltaz / 2, altitude_boundaries.max()
        )

        path_lengths = self.planet.compute_path_length(
            altitude_boundaries, viewer, tangent
        )

        return [l for _, l in path_lengths]

    def path_integral(
        self, wngrid: npt.NDArray[np.float64], return_contrib: t.Optional[bool] = False
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Compute path integral.

        Calculates the absorption and optical depth for each layer assuming
        hemispherical geometry.

        """
        dz = self.deltaz

        total_layers = self.nLayers

        wngrid_size = wngrid.shape[0]

        density_profile = self.densityProfile

        if self.new_method:
            path_length = self.compute_path_length()
        else:
            path_length = self.compute_path_length_old(dz)
        self.path_length = path_length

        tau = np.zeros(shape=(total_layers, wngrid_size), dtype=np.float64)

        for layer in range(total_layers):
            self.debug("Computing layer %s", layer)
            dl = path_length[layer]

            end_k = total_layers - layer

            for contrib in self.contribution_list:
                if tau[layer].min() > 10:
                    break
                self.debug("Adding contribution from %s", contrib.name)
                contrib.contribute(
                    self, 0, end_k, layer, layer, density_profile, tau, path_length=dl
                )

        self.debug("tau %s %s", tau, tau.shape)

        absorption, tau = self.compute_absorption(tau, dz)
        return absorption, tau

    def compute_absorption(
        self, tau: npt.NDArray[np.float64], dz: npt.NDArray[np.float64]
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Compute final absorption and optical depth."""
        tau = np.exp(-tau)
        ap = self.altitudeProfile[:, None]
        pradius = self._planet.fullRadius
        sradius = self._star.radius
        _dz = dz[:, None]

        integral = np.sum((pradius + ap) * (1.0 - tau) * _dz * 2.0, axis=0)
        return ((pradius**2.0) + integral) / (sradius**2), tau

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for this class."""
        return (
            "transmission",
            "transit",
        )



"""Radiative transfer modeeling of eclipses."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.chemistry import Chemistry
from taurex.constants import PI
from taurex.core import derivedparam
from taurex.log import setup_log
from taurex.output import OutputGroup
from taurex.planet import Planet
from taurex.pressure import PressureProfile
from taurex.stellar import Star
from taurex.temperature import TemperatureProfile
from taurex.util.emission import black_body

from .simplemodel import OneDForwardModel

if t.TYPE_CHECKING:
    from taurex.contributions import Contribution
else:
    Contribution = object

_log = setup_log(__name__)


def contribute_ktau_emission_numpy(
    start_k, end_k, density_offset, sigma, density, path, weights, ngrid, layer, ngauss
):
    _path = path[start_k:end_k]
    _density = density[start_k + density_offset : end_k + density_offset]

    _sigma = sigma[start_k + layer : end_k + layer, :, :]

    return np.sum(_sigma * _path * _density, axis=(0, -1))


def contribute_ktau_emission_numba(
    start_k, end_k, density_offset, sigma, density, path, weights, ngrid, layer, ngauss
):
    tau_temp = np.zeros(shape=(ngrid, ngauss))

    for k in range(start_k, end_k):
        _path = path[k]
        _density = density[k + density_offset]
        # for mol in range(nmols):
        for wn in range(ngrid):
            for g in range(ngauss):
                tau_temp[wn, g] += sigma[k + layer, wn, g] * _path * _density
    return tau_temp


try:
    import numba

    contribute_ktau_emission = numba.jit(
        contribute_ktau_emission_numba, nopython=True, nogil=True, fastmath=True
    )
except ImportError:
    _log.warning("Numba not installed, using numpy")
    contribute_ktau_emission = contribute_ktau_emission_numpy


class EmissionModel(OneDForwardModel):
    """A forward model for eclipses."""

    def __init__(
        self,
        planet: t.Optional[Planet] = None,
        star: t.Optional[Star] = None,
        pressure_profile: t.Optional[PressureProfile] = None,
        temperature_profile: t.Optional[TemperatureProfile] = None,
        chemistry: t.Optional[Chemistry] = None,
        nlayers: t.Optional[int] = 100,
        atm_min_pressure: t.Optional[float] = 1e-4,
        atm_max_pressure: t.Optional[float] = 1e6,
        contributions: t.Optional[t.List[Contribution]] = None,
        ngauss: t.Optional[int] = 4,
    ):
        """Initialise eclipse forward model.


        Parameters
        ----------
        name:
            Name to use in logging

        planet:
            Planet model, default planet is Jupiter

        star:
            Star model, default star is Sun-like

        pressure_profile:
            Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
            and ``atm_max_pressure``

        temperature_profile:
            Temperature model, default is an
            :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
            profile at 1500 K

        chemistry:
            Chemistry model, default is
            :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry`
            with ``H2O`` and ``CH4``

        nlayers:
            Number of layers. Used if ``pressure_profile`` is not defined.

        atm_min_pressure:
            Pressure at TOA. Used if ``pressure_profile`` is not defined.

        atm_max_pressure:
            Pressure at BOA. Used if ``pressure_profile`` is not defined.

        ngauss:
            Number of gaussian quadrature points, default = 4

        """
        super().__init__(
            self.__class__.__name__,
            planet,
            star,
            pressure_profile,
            temperature_profile,
            chemistry,
            nlayers,
            atm_min_pressure,
            atm_max_pressure,
            contributions,
        )

        self.set_num_gauss(ngauss)
        self._clamp = 10

    def set_num_gauss(
        self, value: int, coeffs: t.Optional[npt.NDArray[np.float64]] = None
    ) -> None:
        """Set number of gaussian quadrature points.

        Parameters
        ----------
        value: int
            Number of gaussian quadrature points

        coeffs: np.ndarray, optional
            Coefficients for each gaussian quadrature point

        """
        self._ngauss = int(value)

        mu, weight = np.polynomial.legendre.leggauss(self._ngauss)
        self.set_quadratures(mu, weight, coeffs)

    def set_quadratures(
        self,
        mu: npt.NDArray[np.float64],
        weight: npt.NDArray[np.float64],
        coeffs: t.Optional[npt.NDArray[np.float64]] = None,
    ):
        """Set quadrature points.

        Parameters
        ----------
        mu: np.ndarray
            Quadrature points

        weight: np.ndarray
            Quadrature weights

        coeffs: np.ndarray, optional
            Coefficients for each gaussian quadrature point

        """
        self._mu_quads = (mu + 1) / 2
        self._wi_quads = weight / 2
        self._coeffs = coeffs
        if coeffs is None:
            self._coeffs = np.ones(self._ngauss)

    def compute_final_flux(
        self, f_total: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        r"""Compute final flux value.

        For emission this converts the flux emitted to:

        .. math::

            F = \frac{F_{total}}{\F_{star}} \left(\frac{R_{planet}}{R_{star}}\right)^2

        Parameters
        ----------
        f_total: np.ndarray
            Flux emitted by the planet



        """
        star_sed = self._star.spectralEmissionDensity

        self.debug("Star SED: %s", star_sed)
        # quit()
        star_radius = self._star.radius
        planet_radius = self._planet.fullRadius
        self.debug("star_radius %s", self._star.radius)
        self.debug("planet_radius %s", self._star.radius)
        last_flux = ((f_total + self.albedoterm) / star_sed) * (
            planet_radius / star_radius
        ) ** 2

        self.debug("last_flux %s", last_flux)

        return last_flux

    @property
    def albedoterm(self) -> float:
        """Albedo term."""
        return 0.0

    def partial_model(
        self,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
        cutoff_grid: t.Optional[bool] = True,
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]:
        """Evaluate the flux and quandratues but do not compute the final flux."""
        from taurex.util import clip_native_to_wngrid

        self.initialize_profiles()

        native_grid = self.nativeWavenumberGrid
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)
        self._star.initialize(native_grid)

        for contrib in self.contribution_list:
            contrib.prepare(self, native_grid)

        return self.evaluate_emission(native_grid, False)

    def evaluate_emission_ktables(  # noqa: C901
        self, wngrid: npt.NDArray[np.float64], return_contrib: bool
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]:
        """Evaluate emission flux on quadratures using ktables."""
        from taurex.contributions import AbsorptionContribution
        from taurex.util import compute_dz

        dz = compute_dz(self.altitudeProfile)
        total_layers = self.nLayers
        density = self.densityProfile
        wngrid_size = wngrid.shape[0]
        temperature = self.temperatureProfile
        tau = np.zeros(shape=(self.nLayers, wngrid_size))
        surface_tau = np.zeros(shape=(1, wngrid_size))
        layer_tau = np.zeros(shape=(1, wngrid_size))
        tmp_tau = np.zeros(shape=(1, wngrid_size))
        dtau = np.zeros(shape=(1, wngrid_size))

        mol_type = AbsorptionContribution

        non_molecule_absorption = [
            c for c in self.contribution_list if not isinstance(c, mol_type)
        ]
        contrib_types = [type(c) for c in self.contribution_list]
        molecule_absorption = None
        if mol_type in contrib_types:
            molecule_absorption = self.contribution_list[contrib_types.index(mol_type)]

        _mu = 1.0 / self._mu_quads[:, None]
        _w = self._wi_quads[:, None]

        # Do surface first
        # for layer in range(total_layers):
        for contrib in non_molecule_absorption:
            contrib.contribute(
                self, 0, total_layers, 0, 0, density, surface_tau, path_length=dz
            )

        surface_tau = surface_tau * _mu
        if molecule_absorption is not None:
            for idx, m in enumerate(_mu):
                tmp_tau[...] = 0.0
                molecule_absorption.contribute(
                    self, 0, total_layers, 0, 0, density, tmp_tau, path_length=dz * m
                )
                surface_tau[idx] += tmp_tau[0]

        self.debug("density = %s", density[0])
        self.debug("surface_tau = %s", surface_tau)

        planck_term = black_body(wngrid, temperature[0]) / PI

        intensity = planck_term * (np.exp(-surface_tau))

        for layer in range(total_layers):
            layer_tau[...] = 0.0
            dtau[...] = 0.0
            for contrib in non_molecule_absorption:
                contrib.contribute(
                    self,
                    layer + 1,
                    total_layers,
                    0,
                    0,
                    density,
                    layer_tau,
                    path_length=dz,
                )
                contrib.contribute(
                    self, layer, layer + 1, 0, 0, density, dtau, path_length=dz
                )

            k_dtau = None
            k_layer = None
            wg = None
            if molecule_absorption is not None:
                wg = molecule_absorption.weights
                ng = len(wg)
                sigma = molecule_absorption.sigma_xsec

                k_layer = contribute_ktau_emission(
                    layer + 1,
                    total_layers,
                    0,
                    sigma,
                    density,
                    dz,
                    wg,
                    wngrid_size,
                    0,
                    ng,
                )

                k_dtau = contribute_ktau_emission(
                    layer, layer + 1, 0, sigma, density, dz, wg, wngrid_size, 0, ng
                )
                k_dtau += k_layer

            dtau += layer_tau

            self.debug("dtau[%s]=%s", layer, dtau)
            planck_term = black_body(wngrid, temperature[layer]) / PI
            self.debug("planck_term[%s]=%s,%s", layer, temperature[layer], planck_term)

            dtau_calc = np.exp(-dtau * _mu)
            layer_tau_calc = np.exp(-layer_tau * _mu)

            if molecule_absorption is not None:
                dtau_calc *= np.sum(np.exp(-k_dtau * _mu[:, None]) * wg, axis=-1)
                layer_tau_calc *= np.sum(np.exp(-k_layer * _mu[:, None]) * wg, axis=-1)

            intensity += planck_term * (layer_tau_calc - dtau_calc)

            dtau_calc = 0.0
            if dtau.min() < self._clamp:
                dtau_calc = np.exp(-dtau)
            layer_tau_calc = 0.0
            if layer_tau.min() < self._clamp:
                layer_tau_calc = np.exp(-layer_tau)

            if molecule_absorption is not None:
                if k_dtau.min() < self._clamp:
                    dtau_calc *= np.sum(np.exp(-k_dtau) * wg, axis=-1)
                if k_layer.min() < self._clamp:
                    layer_tau_calc *= np.sum(np.exp(-k_layer) * wg, axis=-1)

            _tau = layer_tau_calc - dtau_calc

            if isinstance(_tau, float):
                tau[layer] += _tau
            else:
                tau[layer] += _tau[0]

        self.debug("intensity: %s", intensity)

        return intensity, _mu, _w, tau

    @property
    def usingKTables(self) -> bool:  # noqa: N802
        """Whether ktables are being used."""
        from taurex.cache import GlobalCache

        return GlobalCache()["opacity_method"] == "ktables"

    def evaluate_emission(
        self, wngrid: npt.NDArray[np.float64], return_contrib: bool
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]:
        """Evaluate emission flux on quadratures."""
        if self.usingKTables:
            return self.evaluate_emission_ktables(wngrid, return_contrib)

        dz = self.deltaz

        total_layers = self.nLayers

        density = self.densityProfile

        wngrid_size = wngrid.shape[0]

        temperature = self.temperatureProfile
        tau = np.zeros(shape=(self.nLayers, wngrid_size))
        surface_tau = np.zeros(shape=(1, wngrid_size))

        layer_tau = np.zeros(shape=(1, wngrid_size))

        dtau = np.zeros(shape=(1, wngrid_size))

        # Do surface first
        # for layer in range(total_layers):
        for contrib in self.contribution_list:
            contrib.contribute(
                self, 0, total_layers, 0, 0, density, surface_tau, path_length=dz
            )
        self.debug("density = %s", density[0])
        self.debug("surface_tau = %s", surface_tau)

        planck_term = black_body(wngrid, temperature[0]) / PI

        _mu = 1.0 / self._mu_quads[:, None]
        _w = self._wi_quads[:, None]
        intensity = planck_term * (np.exp(-surface_tau * _mu))

        self.debug("I1_pre %s", intensity)
        # Loop upwards
        for layer in range(total_layers):
            layer_tau[...] = 0.0
            dtau[...] = 0.0
            for contrib in self.contribution_list:
                contrib.contribute(
                    self,
                    layer + 1,
                    total_layers,
                    0,
                    0,
                    density,
                    layer_tau,
                    path_length=dz,
                )
                contrib.contribute(
                    self, layer, layer + 1, 0, 0, density, dtau, path_length=dz
                )
            # for contrib in self.contribution_list:

            self.debug("Layer_tau[%s]=%s", layer, layer_tau)

            dtau += layer_tau

            dtau_calc = 0.0
            if dtau.min() < self._clamp:
                dtau_calc = np.exp(-dtau)
            layer_tau_calc = 0.0
            if layer_tau.min() < self._clamp:
                layer_tau_calc = np.exp(-layer_tau)

            _tau = layer_tau_calc - dtau_calc

            if isinstance(_tau, float):
                tau[layer] += _tau
            else:
                tau[layer] += _tau[0]

            self.debug("dtau[%s]=%s", layer, dtau)
            planck_term = black_body(wngrid, temperature[layer]) / PI
            self.debug("planck_term[%s]=%s,%s", layer, temperature[layer], planck_term)

            dtau_calc = 0.0
            if dtau.min() < self._clamp:
                dtau_calc = np.exp(-dtau * _mu)
            layer_tau_calc = 0.0
            if layer_tau.min() < self._clamp:
                layer_tau_calc = np.exp(-layer_tau * _mu)

            intensity += planck_term * (layer_tau_calc - dtau_calc)

        self.debug("intensity: %s", intensity)

        return intensity, _mu, _w, tau

    def path_integral(
        self, wngrid: npt.NDArray[np.float64], return_contrib: bool
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Evaluate emission flux and integrate quadratures for flux."""
        intensity, _mu, _w, tau = self.evaluate_emission(wngrid, return_contrib)
        self.debug("intensity: %s", intensity)

        flux_total = 2.0 * np.pi * sum(intensity * (_w / _mu))
        self.debug("flux_total %s", flux_total)

        return self.compute_final_flux(flux_total).ravel(), tau

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write model to output group."""
        model = super().write(output)
        model.write_scalar("ngauss", self._ngauss)
        return model

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for this class."""
        return (
            "emission",
            "eclipse",
        )

    @derivedparam(
        param_name="log_F_bol", param_latex=r"log(F$_\{bol\}$)", compute=False
    )
    def logBolometricFlux(self) -> float:  # noqa: N802
        """log10 Flux integrated over all wavelengths (W m-2)."""
        import math

        from scipy.integrate import simps

        intensity, _mu, _w, tau = self.partial_model()

        flux_total = 2.0 * np.pi * sum(intensity * (_w / _mu))

        flux_wl = flux_total[::-1]
        wlgrid = 10000 / self.nativeWavenumberGrid[::-1]

        return math.log10(simps(flux_wl, wlgrid))



import numpy as np


def normalize(v, axis=0):
    norm = np.linalg.norm(v, axis=axis)
    solution = v/norm
    solution[:, norm == 0] = v[:, norm == 0]
    return solution


def sphere_to_cartesian(R, vec, degrees=True):
    result = np.zeros_like(vec)
    _vec = vec[...]
    if degrees:
        _vec[1:] = np.radians(_vec[1:])
    result[0] = (R+_vec[0])*np.sin(_vec[1])*np.cos(_vec[2])
    result[1] = (R+_vec[0])*np.sin(_vec[1])*np.sin(_vec[2])
    result[2] = (R+_vec[0])*np.cos(_vec[1])
    return result

def compute_line_3d(v, t, axis=0):
    """
    Generates a line given two cartesian points
    """
    v = np.array(v)
    t = np.array(t)
    o = v[...]
    u = normalize(t-v, axis=axis)
    return o, u


def parallel_vector(R, alt, max_alt=1e5):
    """
    Generates a viewing and tangent vectors
    parallel to the surface of a sphere
    """

    if not hasattr(alt, '__len__'):
        alt = np.array([alt])
    viewer = np.zeros(shape=(3, len(alt)))
    tangent = np.zeros_like(viewer)
    viewer[0] = -(R+max_alt*2)
    viewer[1] = R+alt
    tangent[1] = R+alt

    return viewer, tangent


def perpendicular_vector(R, max_alt):
    """
    Generates a viewing and tangent vectors
    perpendicular to a sphere of radius R
    """
    return np.array([0, R+max_alt*2, 0]).reshape(3, 1), \
        np.array([0.0, 0, 0]).reshape(3, 1)

def multi_dot(a, b):
    return np.sum(a*b, axis=0)

def compute_intersection_3d(R, h, u, o, c=np.zeros(3), allow_single=False):
    """
    Computes line-sphere intersection for a range of radius offsets
    Will detect if path crosses R and cutoff the point

    Parameters
    ----------
    R: float
        Radius

    h: float or array_like of shape (m)
        height above radius
    
    u: array with shape (3,n)
        normalized direction vector

    o: array with shape (3,n)
        origin vector

    c: vector, optional
        center position of sphere default is (0,0,0)

    Returns
    --------
    solution: array of shape (2,3,m,n)
        Array containing points that intersect the sphere
        First index is always the point closest to origin vector


    """
    if not hasattr(h, '__len__'):
        h = np.array([h])
    else:
        h = np.array(h)
    
    if len(u.shape) == 1:
        u = u.reshape(-1, 1)
        o = o.reshape(-1, 1)
    
    single_dot = multi_dot(u, o)
    dot_res = (single_dot)**2 - (o**2).sum(axis=0)
    delta = dot_res + (R+h[..., None])**2
    solution = np.zeros(shape=(2, 3, h.shape[0], u.shape[1]))
    sol1 = np.zeros(shape=(3, h.shape[0], u.shape[1]))
    sol2 = np.zeros(shape=(3, h.shape[0], u.shape[1]))
    solution[...] = np.nan
    filt = delta > 0
    if allow_single:
        filt = delta >= 0
    if filt.sum() == 0:
        return None
    
    with np.errstate(divide='ignore', invalid='ignore'):

        d1 = -(single_dot) + np.sqrt(delta)
        d1[d1 < 0] = 0.0
        sol1 = o[:, None, :] + d1*u[:, None]

        d2 = -(single_dot) - np.sqrt(delta)
        d2[d2 < 0] = 0.0
        sol2 = o[:, None, :] + d2*u[:, None]
        v1 = np.sum((o[:, None, :]-sol1)**2, axis=0)
        v2 = np.sum((o[:, None, :]-sol2)**2, axis=0)
    #
    #print(v1.shape)
    max_filter = v2 > v1
    #print(max_filter.shape)
    a_filter = max_filter
    b_filter = ~max_filter
    #print(solution.shape)
    if a_filter.sum() > 0:
        solution[0, :, a_filter] = sol1[:, a_filter].T
        solution[1, :, a_filter] = sol2[:, a_filter].T
    if b_filter.sum() > 0:
        solution[0, :, b_filter] = sol2[:, b_filter].T

        solution[1, :, b_filter] = sol1[:, b_filter].T
    
    # Detect planet crossings
    delta = dot_res + (R)**2

    filt = delta > 0
    tang = np.where(filt)[0]
    if filt.sum() > 0:
        d1 = -(single_dot[filt]) + np.sqrt(delta[filt])
        sol1 = o[:, filt] + d1*u[:, filt]
        d2 = -(single_dot[filt]) - np.sqrt(delta[filt])
        sol2 = o[:, filt] + d2*u[:, filt]
        v1 = ((o[:, filt]-sol1)**2).sum()
        v2 = ((o[:, filt]-sol2)**2).sum()
        if v2 > v1:
            solution[1, ..., tang] = sol1.T[..., None]
        else:
            solution[1, ..., tang] = sol2.T[..., None]

    return solution


def compute_path_length_3d(R, altitudes, viewer, tangent,
                           coordinates='cartesian'):
    """
    Given a viewing and tangent vector, computes the path length for a sphere
    """

    _viewer = viewer[...]
    _tangent = tangent[...]

    if not hasattr(altitudes, '__len__'):
        altitudes = np.array([altitudes])
    else:
        altitudes = np.array(altitudes)

    if not isinstance(coordinates, (list, tuple,)):

        coordinates = [coordinates, coordinates]

    if coordinates[0] in ('spherical'):
        _viewer = sphere_to_cartesian(R, _viewer)
    if coordinates[1] in ('spherical'):
        _tangent = sphere_to_cartesian(R, _tangent)

    if len(_viewer.shape) == 1:
        _viewer = _viewer.reshape(-1, 1)
        _tangent = _viewer.reshape(-1, 1)

    o, u = compute_line_3d(_viewer,
                           _tangent)
    intersections = compute_intersection_3d(R, altitudes, u, o)

    if intersections is not None:

        distances = (np.linalg.norm(intersections[1] - intersections[0],
                                    axis=0))
        filt = np.isfinite(distances)
        all_distances = []

        for i in range(_viewer.shape[1]):
            layer_filt = filt[:, i]
            good_indices = np.where(layer_filt)[0]
            dists = distances[layer_filt, i]
            final_distances = np.zeros_like(dists)
            final_distances[0] = dists[0]
            final_distances[1:] = dists[1:]-dists[:-1]

            all_distances.append((good_indices, final_distances))
        return all_distances
    else:
        return None


"""Modules that deal with computing contributions to optical depth."""


from .absorption import AbsorptionContribution
from .cia import CIAContribution, contribute_cia
from .contribution import Contribution, contribute_tau
from .flatmie import FlatMieContribution
from .hm import HydrogenIon
from .leemie import LeeMieContribution
from .rayleigh import RayleighContribution
from .simpleclouds import SimpleCloudsContribution

__all__ = [
    "Contribution",
    "AbsorptionContribution",
    "CIAContribution",
    "RayleighContribution",
    "SimpleCloudsContribution",
    "LeeMieContribution",
    "FlatMieContribution",
    "HydrogenIon",
    "contribute_tau",
    "contribute_cia",
]



# flake8: noqa

"""Lightcurve model class.

This will probably be replaced by a more general forward model class
in the future.

"""
import pickle  # noqa: S403
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.binning.lightcurvebinner import LightcurveBinner
from taurex.chemistry import Chemistry
from taurex.data.fittable import fitparam
from taurex.model import ForwardModel, OneDForwardModel
from taurex.pressure import PressureProfile
from taurex.types import PathLike

from .lightcurvedata import InstrumentType, LightCurveData


class LightCurveModel(ForwardModel):
    """A base class for producing forward models"""

    @property
    def defaultBinner(self) -> LightcurveBinner:  # noqa: N802
        """Default binner for lightcurve."""
        return LightcurveBinner

    def __init__(
        self,
        forward_model: OneDForwardModel,
        file_loc: PathLike,
        instruments: t.Optional[
            t.Union[t.Sequence[InstrumentType], t.Literal["all"]]
        ] = None,
    ) -> None:
        super().__init__("LightCurveModel")

        self._forward_model = forward_model
        self.file_loc = file_loc
        self._load_file()

        self._load_orbital_profile()

        if instruments is None:
            instruments = "all"

        if instruments == "all" or "all" in instruments:
            instruments = LightCurveData.availableInstruments

        self._load_instruments(instruments)
        self._load_ldcoeff()
        self._initialize_lightcurves()
        self.setup_binner()

    def setup_binner(self) -> None:
        """Setup binner for the observed."""
        from taurex.data.spectrum import ArraySpectrum

        self._wngrid = 10000 / self.lc_data["obs_spectrum"][:, 0]
        self._arr_spec = ArraySpectrum(self.lc_data["obs_spectrum"])
        self._binner = self._arr_spec.create_binner()

    def initialize_profiles(self) -> None:
        """Initialize profiles."""
        self._forward_model.initialize_profiles()

    def _load_file(self) -> None:
        """Load lightcurve data from pickle file."""
        # input data from lightcurve, not from spectrum.

        with open(self.file_loc, "rb") as f:
            self.lc_data = pickle.load(f, encoding="latin1")  # noqa: S301

    def _load_orbital_profile(self) -> None:
        """Load orbital information"""
        self.info("Load lc information")

        # new
        self.mid_time = self.lc_data["orbital_info"]["mt"]
        self._inclination = self.lc_data["orbital_info"]["i"]
        self.period = self.lc_data["orbital_info"]["period"]
        self.periastron = self.lc_data["orbital_info"]["periastron"]
        self.sma_over_rs = self.lc_data["orbital_info"]["sma_over_rs"]
        self.ecc = self.lc_data["orbital_info"]["e"]

    def _load_ldcoeff(self) -> None:
        """Load limb-darkening coefficients"""
        self.ld_coeff_file = np.array([])
        for i in self._instruments:
            self.ld_coeff_file = np.append(
                self.ld_coeff_file, self.lc_data[i.instrumentName]["ld_coeff"]
            ).reshape(-1, 4)
        if np.shape(self.ld_coeff_file)[1] == 4:
            raise ValueError("please use 4 ldcoeff law.")

    def _load_instruments(self, instruments: t.Sequence[InstrumentType]) -> None:
        """Load instruments from lightcurve data."""
        self._instruments: t.List[LightCurveData] = []
        # new
        ins_keys = self.lc_data.keys()
        for ins in instruments:
            if ins in ins_keys:
                self.info(f"Loading {ins} light curves")
                self._instruments.append(
                    LightCurveData.fromInstrumentName(ins, self.lc_data)
                )
            else:
                self.info(f"Could not find {ins} in instrument keys")
        new_instrument_list = sorted(
            self._instruments, key=lambda x: x.wavelengthRegion[0], reverse=True
        )
        self._instruments = new_instrument_list

    def _initialize_lightcurves(self) -> None:
        """Initialize lightcurves."""
        nFactor = [ins.minNFactors for ins in self._instruments]

        min_n_factors = None

        max_n_factors = None
        if len(nFactor) > 0:
            min_n_factors = np.concatenate(nFactor)

        nFactor = [ins.maxNFactors for ins in self._instruments]
        if len(nFactor) > 0:
            max_n_factors = np.concatenate(nFactor)

        if max_n_factors is not None and min_n_factors is not None:
            self.create_normalization_fitparams()

            for idx, value in enumerate(zip(min_n_factors, max_n_factors)):
                min_n, max_n = value
                self.modify_bounds(f"Nfactor_{idx}", [min_n, max_n])

        min_time = min([ins.timeSeries.min() for ins in self._instruments])

        max_time = max([ins.timeSeries.max() for ins in self._instruments])

        self.modify_bounds("mid_transit_time", [min_time, max_time])

    @fitparam(
        param_name="sma_over_rs",
        param_latex="sma_over_rs",
        default_mode="linear",
        default_fit=False,
        default_bounds=[1.001, 60.0],
    )
    def semiMajorAxisOverRs(self) -> float:  # noqa: N802
        """Semi-major axis over stellar radius."""
        return self.sma_over_rs

    @semiMajorAxisOverRs.setter
    def semiMajorAxisOverRs(self, value: float) -> None:  # noqa: N802
        """Set semi-major axis over stellar radius."""
        self.sma_over_rs = value

    @fitparam(
        param_name="mid_transit_time",
        param_latex="mid_transit_time",
        default_mode="linear",
        default_fit=False,
        default_bounds=[1.001, 2.0],
    )
    def midTransitTime(self) -> float:  # noqa: N802
        """Mid-transit time."""
        return self.mid_time

    @midTransitTime.setter
    def midTransitTime(self, value: float) -> None:  # noqa: N802
        """Set mid-transit time."""
        self.mid_time = value

    @fitparam(
        param_name="inclination",
        param_latex="inclination",
        default_mode="linear",
        default_fit=False,
        default_bounds=[40.0, 90.0],
    )
    def inclination(self) -> float:
        """Inclination."""
        return self._inclination

    @inclination.setter
    def inclination(self, value: float) -> None:
        """Set inclination."""
        self._inclination = value

    @property
    def temperatureProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature profile."""
        return self._forward_model.temperatureProfile

    @property
    def pressureProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure profile."""
        return self._forward_model.pressureProfile

    @property
    def densityProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Density profile."""
        return self._forward_model.densityProfile

    @property
    def scaleheight_profile(self) -> npt.NDArray[np.float64]:
        """Scaleheight profile."""
        return self._forward_model.scaleheight_profile

    @property
    def chemistry(self) -> Chemistry:
        """Chemistry."""
        return self._forward_model.chemistry

    @property
    def gravity_profile(self) -> npt.NDArray[np.float64]:
        """Gravity profile."""
        return self._forward_model.gravity_profile

    @property
    def pressure(self) -> PressureProfile:
        """Pressure."""
        return self._forward_model.pressure

    @property
    def altitudeProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Altitude profile."""
        return self._forward_model.altitudeProfile

    def create_normalization_fitparams(self):
        """Create normalization fitparams for each instrument."""
        import itertools

        ins_name = list(
            itertools.chain(
                *tuple(
                    [
                        [ins.instrumentName] * ins.minNFactors.shape[0]
                        for ins in self._instruments
                    ]
                )
            )
        )

        ins_number = list(
            itertools.chain(
                *tuple(
                    [list(range(ins.minNFactors.shape[0])) for ins in self._instruments]
                )
            )
        )

        for idx, val in enumerate(zip(ins_name, ins_number)):
            name, no = val
            param_name = f"Nfactor_{idx}"
            param_latex = f"{name}_{no}"

            default_fit = False
            default_bounds = [0, 1]
            default_mode = "linear"

            def read_n(self, idx=idx):
                return self._nfactor[idx]

            def write_n(self, value, idx=idx):
                self._nfactor[idx] = value

            read_n.__doc__ = f"""Read Nfactor for {idx}"""

            self.add_fittable_param(
                param_name,
                param_latex,
                read_n,
                write_n,
                default_mode,
                default_fit,
                default_bounds,
            )

    def instrument_light_curve(
        self, model: npt.NDArray[np.float64], wlgrid: npt.NDArray[np.float64]
    ):
        """Combine light-curves from different instrucments together."""
        result = np.array([])
        sma_over_rs_value = self.sma_over_rs
        inclination_value = self.inclination
        mid_time_value = self.mid_time
        try:
            Nfactor = self._nfactor
        except AttributeError:
            Nfactor = np.ones_like(wlgrid)

        result = []

        for ins in self._instruments:
            min_wl, max_wl = ins.wavelengthRegion
            index = (wlgrid > min_wl) & (wlgrid < max_wl)

            lc = self.light_curve_chain(
                model[index],
                time_array=ins.timeSeries,
                period=self.period,
                sma_over_rs=sma_over_rs_value,
                eccentricity=self.ecc,
                inclination=inclination_value,
                periastron=self.periastron,
                mid_time=mid_time_value,
                ldcoeff=self.ld_coeff_file[index],
                Nfactor=Nfactor[index],
            )
            result.append(lc)

        return np.concatenate(result)

    def light_curve_chain(
        self,
        model,
        time_array,
        period,
        sma_over_rs,
        eccentricity,
        inclination,
        periastron,
        mid_time,
        ldcoeff,
        Nfactor,
    ):
        """Create model light-curve and lightcurve chain."""
        import pylightcurve as plc

        from taurex.model import EmissionModel, TransmissionModel

        result = []
        sqrt_model = np.sqrt(model)
        # self.info('Creating Lightcurve chain.')
        for n in range(len(sqrt_model)):
            if isinstance(self._forward_model, TransmissionModel):
                self.debug("Using Transit")
                transit_light_curve = plc.transit(
                    "claret",
                    ldcoeff[n],
                    sqrt_model[n],
                    period,
                    sma_over_rs,
                    eccentricity,
                    inclination,
                    periastron,
                    mid_time,
                    time_array,
                )
                result.append(transit_light_curve * Nfactor[n])
            elif isinstance(self._forward_model, EmissionModel):
                self.debug("Using Eclipse")
                rp_over_rs = (
                    self._forward_model.planet.fullRadius
                    / self._forward_model.star.radius
                )
                self.debug("rp_over_rs %s", rp_over_rs)
                self.debug("fp_over_fs %s", sqrt_model)

                eclipse_light_curve = plc.eclipse(
                    sqrt_model[n],
                    rp_over_rs,
                    period,
                    sma_over_rs,
                    eccentricity,
                    inclination,
                    periastron,
                    mid_time,
                    time_array,
                )
                result.append(eclipse_light_curve * Nfactor[n])

        self.debug("Result %s", result)
        return np.concatenate(result)

    def build(self):
        self._forward_model.build()

        self._fitting_parameters = {}
        self._fitting_parameters.update(self.fitting_parameters())

        self._fitting_parameters.update(self._forward_model.fittingParameters)

        self._derived_parameters = {}
        self._derived_parameters.update(self.derived_parameters())
        self._derived_parameters.update(self._forward_model.derivedParameters)

    @property
    def nativeWavenumberGrid(self):
        return self._forward_model.nativeWavenumberGrid

    def model(self, wngrid=None, cutoff_grid=True):
        """Computes the forward model for a wngrid"""
        # new

        native_grid, model, tau, extra = self._forward_model.model(
            self._wngrid, cutoff_grid
        )
        binned_model = self._binner.bindown(native_grid, model)
        wlgrid = 10000 / self._wngrid
        result = self.instrument_light_curve(binned_model[1], wlgrid)

        return self._wngrid, result, tau, [native_grid, model, binned_model[1], extra]

    def compute_error(self, samples, wngrid=None, binner=None):
        from taurex.util.math import OnlineVariance

        tp_profiles = OnlineVariance()
        active_gases = OnlineVariance()
        inactive_gases = OnlineVariance()

        lc_spectrum = OnlineVariance()

        native_spectrum = OnlineVariance()
        binned_spectrum = OnlineVariance()

        for weight in samples():
            grid, lc, tau, extra = self.model(wngrid=wngrid, cutoff_grid=False)
            native_grid, native, binned, _ = extra

            tp_profiles.update(self.temperatureProfile, weight=weight)
            active_gases.update(self.chemistry.activeGasMixProfile, weight=weight)
            inactive_gases.update(self.chemistry.inactiveGasMixProfile, weight=weight)

            native_spectrum.update(native, weight=weight)
            binned_spectrum.update(binned, weight=weight)
            lc_spectrum.update(lc, weight=weight)

        profile_dict = {}
        spectrum_dict = {}

        tp_std = np.sqrt(tp_profiles.parallelVariance())
        active_std = np.sqrt(active_gases.parallelVariance())
        inactive_std = np.sqrt(inactive_gases.parallelVariance())

        profile_dict["temp_profile_std"] = tp_std
        profile_dict["active_mix_profile_std"] = active_std
        profile_dict["inactive_mix_profile_std"] = inactive_std

        spectrum_dict["native_std"] = np.sqrt(native_spectrum.parallelVariance())
        spectrum_dict["binned_std"] = np.sqrt(binned_spectrum.parallelVariance())
        spectrum_dict["lightcurve_std"] = np.sqrt(lc_spectrum.parallelVariance())
        return profile_dict, spectrum_dict

    def model_contrib(self, wngrid=None, cutoff_grid=True):
        # new
        wngrid = 10000 / self.lc_data["obs_spectrum"][:, 0]
        native_grid, contribs = self._forward_model.model_contrib(wngrid, cutoff_grid)
        binner = self._binner
        all_contrib_dict = {}

        wlgrid = 10000 / wngrid

        for contrib_name, main_contribution in contribs.items():
            model, tau, extras = main_contribution
            binned = binner.bindown(native_grid, model)[1]

            result = self.instrument_light_curve(binned, wlgrid)
            all_contrib_dict[contrib_name] = (
                result,
                tau,
                [native_grid, model, binned, extras],
            )

        return native_grid, all_contrib_dict

    def model_full_contrib(self, wngrid=None, cutoff_grid=True):
        """Computes the forward model for a wngrid for each contribution"""

        # new
        wngrid = 10000 / self.lc_data["obs_spectrum"][:, 0]

        native_grid, contrib_res = self._forward_model.model_full_contrib(
            wngrid, cutoff_grid
        )
        binner = self._binner

        self.info("Computing lightcurve contribution")
        wlgrid = 10000 / wngrid
        lc_contrib_res = {}

        # Loop through each contribtuion
        for contrib_name, contrib_list in contrib_res.items():
            lc_contrib_list = []

            for name, native, tau, extra in contrib_list:
                binned = binner.bindown(native_grid, native)[1]
                result = self.instrument_light_curve(binned, wlgrid)

                new_packed = name, result, tau, (native_grid, native, binned, extra)

                lc_contrib_list.append(new_packed)

            lc_contrib_res[contrib_name] = lc_contrib_list

        return native_grid, lc_contrib_res

    def write(self, output):
        lc = output.create_group("Lightcurve")

        lc_grps = lc.create_group("Instrument")
        for ins in self._instruments:
            ins.write(lc_grps)

        lc.write_scalar("mid_time", self.mid_time)
        lc.write_scalar("inclination", self._inclination)
        lc.write_scalar("period", self.period)
        lc.write_scalar("periastron", self.periastron)
        lc.write_scalar("sma_over_rs", self.sma_over_rs)
        lc.write_scalar("eccentricity", self.ecc)

        self._forward_model.write(output)

        return None




def getattr_recursive(obj, attr):
    split = attr.split('.')
    if len(split) == 1:
        return getattr(obj, split[0])
    else:
        return getattr_recursive(getattr(obj, split[0]), '.'.join(split[1:]))


def setattr_recursive(obj, attr, value):
    split = attr.split('.')
    if len(split) == 1:
        setattr(obj, split[0], value)
    else:
        setattr_recursive(getattr(obj, split[0]), '.'.join(split[1:]), value)


def runfunc_recursive(obj, func, *args, **kwargs):
    split = func.split('.')
    if len(split) == 1:
        return getattr(obj, split[0])(*args, **kwargs)
    else:
        return runfunc_recursive(getattr(obj, split[0]), '.'.join(split[1:]),
                                 *args, **kwargs)



import pathlib
import typing as t

import configobj

from taurex.cache import GlobalCache
from taurex.log import Logger
from taurex.optimizer import Optimizer
from taurex.types import PathLike

from .factory import (
    create_chemistry,
    create_instrument,
    create_model,
    create_observation,
    create_optimizer,
    create_planet,
    create_pressure_profile,
    create_star,
    create_temperature_profile,
)


class ParameterParser(Logger):
    """Parse input file and generate appropriate objects."""

    def __init__(self):
        super().__init__("ParamParser")
        self._read = False

    def transform(self, section: t.Dict[str, t.Any], key: str) -> t.Any:
        """Transform string arguments to appropriate types."""
        val = section[key]
        newval = val
        if isinstance(val, list):
            try:
                newval = list(map(float, val))
            except Exception:
                return newval
        elif isinstance(val, (str)):
            if val.lower() in [
                "true",
                "yes",
                "yeah",
                "yup",
                "certainly",
                "uh-huh",
            ]:
                newval = True
            elif val.lower() in [
                "false",
                "no",
                "nope",
                "no-way",
                "hell-no",
            ]:
                newval = False
            else:
                try:
                    newval = float(val)
                except Exception:
                    return newval
        section[key] = newval
        return newval

    def setup_globals(self):  # noqa: C901
        """Setup global cache from input file."""
        from taurex.cache import CIACache, GlobalCache, OpacityCache

        config = self._raw_config.dict()

        if "Global" in config:
            global_config = config["Global"]
            try:
                OpacityCache().set_opacity_path(config["Global"]["xsec_path"])
            except KeyError:
                self.warning("No xsec path set, opacities " "cannot be used in model")
            try:
                OpacityCache().set_interpolation(config["Global"]["xsec_interpolation"])
                self.info(
                    "Interpolation mode set "
                    "to {}".format(config["Global"]["xsec_interpolation"])
                )
            except KeyError:
                self.info("Interpolation mode set to linear")

            try:
                CIACache().set_cia_path(config["Global"]["cia_path"])
            except KeyError:
                self.warning("No cia path set, cia cannot be used in model")

            try:
                OpacityCache().set_memory_mode(config["Global"]["xsec_in_memory"])
            except KeyError:
                self.warning("Xsecs will be loaded in memory")

            try:
                OpacityCache().enable_radis(config["Global"]["use_radis"])
            except KeyError:
                self.warning("Radis is disabled")

            try:
                wn_start, wn_end, wn_points = config["Global"]["radis_grid"]

                OpacityCache().set_radis_wavenumber(wn_start, wn_end, wn_points)
            except KeyError:
                self.warning("Radis default grid will be used")

            try:
                extension_paths = config["Global"]["extension_paths"]
                if isinstance(extension_paths, str):
                    extension_paths = [
                        extension_paths,
                    ]

                from .classfactory import ClassFactory

                ClassFactory().set_extension_paths(paths=extension_paths)
            except KeyError:
                pass

            gc = GlobalCache()

            for key, value in global_config.items():
                gc[key] = value

    def read(self, filename: PathLike) -> None:
        """Read input file into a dictionary and transform."""
        path = pathlib.Path(filename)
        if not path.exists():
            raise Exception(f"Input file {filename} does not exist")
        self._raw_config = configobj.ConfigObj(filename)
        self.debug(
            "Raw Config file is {}, filename "
            "is {}".format(self._raw_config, filename)
        )
        self._raw_config.walk(self.transform)
        config = self._raw_config.dict()
        self.debug("Config file is {}, filename " "is {}".format(config, filename))

    def generate_lightcurve(self):
        """Generate lightcurve model from input file."""
        config = self._raw_config.dict()
        if "Lightcurve" in config:
            from taurex.model.lightcurve.lightcurve import LightCurveModel

            model = self.generate_model()
            lightcurvefile = config["Lightcurve"]["lc_pickle"]
            return LightCurveModel(model, lightcurvefile)
        else:
            raise KeyError

    def generate_appropriate_model(self, obs=None):
        """Generate appropriate model from input file."""
        try:
            return self.generate_lightcurve()
        except KeyError:
            pass
        return self.generate_model(obs=obs)

    def generate_instrument(self, binner=None):
        """Generate instrument from input file."""
        from taurex.binning import NativeBinner

        config = self._raw_config.dict()
        if "Instrument" in config:
            inst_config = config["Instrument"]

            try:
                num_obs = inst_config.pop("num_observations")
            except KeyError:
                num_obs = 1

            if "instrument" in inst_config:
                _bin = binner or NativeBinner
                if inst_config["instrument"].lower() in (
                    "snr",
                    "signalnoise",
                ):
                    inst = self.create_snr(_bin, inst_config)
                    return inst, num_obs
            inst = create_instrument(inst_config)
            return inst, num_obs
        else:
            return None

    def create_snr(self, binner, config):
        """Create SNR instrument."""
        from taurex.instruments.snr import SNRInstrument

        if binner is None:
            self.critical("Binning must be defined for SNR instrument")
            raise ValueError("Binning must be defined for SNR instrument")
        else:
            snr = 10
            if "SNR" in config:
                snr = config["SNR"]

            return SNRInstrument(SNR=snr, binner=binner)

    def generate_optimizer(self):
        """Generate optimizer from input file."""
        config = self._raw_config.dict()
        if "Optimizer" in config:
            return create_optimizer(config["Optimizer"])

    def setup_optimizer(self, optimizer: Optimizer):
        """Setup fitting parameters for optimizer."""
        fitting_parameters = self.generate_fitting_parameters()

        self.info("Setting up optimizer")

        for key, value in fitting_parameters.items():
            fit = value["fit"]
            bounds = value["bounds"]
            mode = value["mode"]
            factor = value["factor"]
            prior = value["prior"]

            if fit:
                self.info("Fitting: %s", key)
                optimizer.enable_fit(key)
            else:
                optimizer.disable_fit(key)

            if factor:
                optimizer.set_factor_boundary(key, factor)

            if bounds:
                optimizer.set_boundary(key, bounds)

            if mode:
                optimizer.set_mode(key, mode.lower())

            if prior is not None:
                optimizer.set_prior(key, prior)

        fitting_parameters = self.generate_derived_parameters()

        for key, value in fitting_parameters.items():
            compute = value["compute"]

            if compute is not None:
                if compute:
                    self.info("Deriving %s", key)
                    optimizer.enable_derived(key)
                else:
                    optimizer.disable_derived(key)

    def generate_observation(self):
        """Generate observation from input file."""
        config = self._raw_config.dict()
        if "Observation" in config:
            observation_config = config["Observation"]
            if "lightcurve" in observation_config:
                from taurex.data.spectrum.lightcurve import ObservedLightCurve

                return ObservedLightCurve(observation_config["lightcurve"])

            elif "observed_spectrum" in observation_config:
                from taurex.data.spectrum.observed import ObservedSpectrum

                return ObservedSpectrum(observation_config["observed_spectrum"])

            elif "taurex_spectrum" in observation_config:
                if observation_config["taurex_spectrum"] == "self":
                    return "self"
                from taurex.data.spectrum.taurex import TaurexSpectrum

                return TaurexSpectrum(observation_config["taurex_spectrum"])
            elif "iraclis_spectrum" in observation_config:
                from taurex.data.spectrum.iraclis import IraclisSpectrum

                return IraclisSpectrum(observation_config["iraclis_spectrum"])
            else:
                config = self._raw_config.dict()
                return create_observation(observation_config)
        return None

    def create_manual_binning(self, config):
        import math

        import numpy as np

        from taurex.binning import FluxBinner, SimpleBinner

        binning_class = SimpleBinner

        if "accurate" in config:
            if config["accurate"]:
                binning_class = FluxBinner

        # Handle wavelength grid
        wngrid = None
        if "wavelength_grid" in config:
            start, end, size = config["wavelength_grid"]
            wngrid = 10000 / np.linspace(start, end, int(size))
            wngrid = np.sort(wngrid)

        elif "wavenumber_grid" in config:
            start, end, size = config["wavenumber_grid"]
            wngrid = np.linspace(start, end, int(size))

        elif "log_wavenumber_grid" in config:
            start, end, size = config["log_wavenumber_grid"]
            start = math.log10(start)
            end = math.log10(end)
            wngrid = np.logspace(start, end, int(size))

        elif "log_wavelength_grid" in config:
            start, end, size = config["log_wavelength_grid"]
            start = math.log10(start)
            end = math.log10(end)
            wngrid = np.sort(10000 / np.logspace(start, end, int(size)))
        elif "wavelength_res" in config:
            from taurex.util import create_grid_res

            start, end, res = config["wavelength_res"]
            wlgrid = create_grid_res(res, start, end)[:, 0].flatten()
            wngrid = 10000 / wlgrid[::-1]

        if wngrid is None:
            self._logger.error(
                "manual was selected and no grid was given."
                "(Use wavelength_grid, wavenumber_grid or "
                "log versions)"
            )
            raise Exception("manual selected but no grid given")

        return binning_class(wngrid), wngrid

    def generate_binning(self):
        config = self._raw_config.dict()
        if "Binning" in config:
            binning_config = config["Binning"]
            if "bin_type" in binning_config:
                bin_type = binning_config["bin_type"].lower()

                if bin_type == "native":
                    return "native"
                elif bin_type == "observed":
                    return "observed"
                elif bin_type == "manual":
                    return self.create_manual_binning(binning_config)
            else:
                return None
        return None

    def generate_model(
        self,
        chemistry=None,
        pressure=None,
        temperature=None,
        planet=None,
        star=None,
        obs=None,
    ):
        config = self._raw_config.dict()
        if "Model" in config:
            if chemistry is None:
                chemistry = self.generate_chemistry_profile()
            if pressure is None:
                pressure = self.generate_pressure_profile()
            if temperature is None:
                temperature = self.generate_temperature_profile()
            if planet is None:
                planet = self.generate_planet()
            if star is None:
                star = self.generate_star()
            model = create_model(
                config["Model"],
                chemistry,
                temperature,
                pressure,
                planet,
                star,
                observation=obs,
            )
        else:
            return None

        return model

    def generate_chemistry_profile(self):
        config = self._raw_config.dict()
        if "Chemistry" in config:
            return create_chemistry(config["Chemistry"])
        else:
            return None

    def generate_pressure_profile(self):
        config = self._raw_config.dict()
        if "Pressure" in config:
            return create_pressure_profile(config["Pressure"])
        else:
            return None

    def generate_temperature_profile(self):
        config = self._raw_config.dict()
        if "Temperature" in config:
            return create_temperature_profile(config["Temperature"])
        else:
            return None

    def generate_planet(self):
        config = self._raw_config.dict()

        if "Planet" in config:
            return create_planet(config["Planet"])
        else:
            return None

    def generate_star(self):
        config = self._raw_config.dict()

        if "Star" in config:
            return create_star(config["Star"])
        else:
            return None

    def generate_fitting_parameters(self):
        from .factory import create_prior

        config = self._raw_config.dict()
        if "Fitting" in config:
            fitting_config = config["Fitting"]

            fitting_params = {}

            for key, value in fitting_config.items():
                fit_param, fit_type = key.split(":")
                if fit_param not in fitting_params:
                    fitting_params[fit_param] = {
                        "fit": False,
                        "bounds": None,
                        "mode": None,
                        "factor": None,
                        "prior": None,
                    }

                if fit_type == "prior":
                    value = create_prior(value)
                fitting_params[fit_param][fit_type] = value

            if GlobalCache()["implicit_fit"]:
                for value in fitting_params.values():
                    value["fit"] = True
            return fitting_params

        else:
            return {}

    def generate_derived_parameters(self):
        config = self._raw_config.dict()
        if "Derive" in config:
            fitting_config = config["Derive"]

            fitting_params = {}

            for key, value in fitting_config.items():
                fit_param, fit_type = key.split(":")
                if fit_param not in fitting_params:
                    fitting_params[fit_param] = {"compute": None}
                fitting_params[fit_param][fit_type] = value

            return fitting_params
        else:
            return {}



"""Defines constants used in taurex"""
import astropy.constants as c
import numpy as np

from taurex.util import conversion_factor

AMU = conversion_factor("u", "kg")
KBOLTZ = c.k_B.value
G = c.G.value
RSOL = conversion_factor("Rsun", "m")
RJUP = conversion_factor("Rjup", "m")
PI = np.pi
MSOL = conversion_factor("Msun", "kg")

MJUP = conversion_factor("Mjup", "kg")
AU = conversion_factor("AU", "m")
PLANCK = c.h.value
SPDLIGT = conversion_factor("c", "m/s")


def get_constant(name, unit=None):
    from taurex.util import conversion_factor

    const = getattr(c, name)
    base_unit = str(const.unit)
    const = const.value
    if unit is not None:
        const *= conversion_factor(base_unit, unit)
    return const



"""These modules deal with binning down results from models."""
from .binner import BinDownType, BinnedSpectrumType, Binner
from .fluxbinner import FluxBinner
from .nativebinner import NativeBinner
from .simplebinner import SimpleBinner

__all__ = [
    "Binner",
    "BinDownType",
    "BinnedSpectrumType",
    "SimpleBinner",
    "FluxBinner",
    "NativeBinner",
]



"""Opacity loaded from HDF5 file."""

import pathlib
import typing as t

import numpy as np
import numpy.typing as npt
from astropy.units import UnitConversionError

from taurex.cache import GlobalCache
from taurex.mpi import allocate_as_shared
from taurex.types import PathLike

from .interpolateopacity import InterpModeType, InterpolatingOpacity


class HDF5Opacity(InterpolatingOpacity):
    """Opacities from HDF5 files.

    These usually come from ExoMol. Specifically the ExoMolOP database.

    """

    @classmethod
    def priority(cls) -> int:
        return 5

    @classmethod
    def discover(
        cls,
    ) -> t.List[t.Tuple[str, t.Tuple[pathlib.Path, InterpModeType, bool]]]:
        """Discover opacities from HDF5 files in path."""
        from taurex.cache import GlobalCache

        path = GlobalCache()["xsec_path"]
        if path is None:
            return []

        path = pathlib.Path(path)

        file_list = list(path.glob("*.h5")) + list(path.glob("*.hdf5"))

        discovery = []

        interp = GlobalCache()["xsec_interpolation"] or "linear"
        mem = GlobalCache()["xsec_in_memory"] or True

        for f in file_list:
            op = HDF5Opacity(f, interpolation_mode="linear", in_memory=False)
            mol_name = op.moleculeName
            discovery.append((mol_name, [f, interp, mem]))
            # op._spec_dict.close()
            del op

        return discovery

    def __init__(
        self,
        filename: PathLike,
        interpolation_mode: t.Optional[InterpModeType] = "exp",
        in_memory: t.Optional[bool] = False,
    ) -> None:
        """Initialize opacity from HDF5 file.

        Parameters
        ----------
        filename: PathLike
            Path to opacity file.

        interpolation_mode: str
            Interpolation mode. Either 'linear' or 'exp'.

        in_memory: bool
            If True, the opacity will be loaded in shared memory

        Raises
        ------
        FileNotFoundError
            If file does not exist.

        """
        filename = pathlib.Path(filename)
        super().__init__(
            f"HDF5Opacity:{filename.stem[:10]}",
            interpolation_mode=interpolation_mode,
        )

        if not filename.exists():
            raise FileNotFoundError(f"Could not find {filename}")

        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self.in_memory = in_memory
        self._molecular_citation = []
        self._load_hdf_file(filename)

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Name of molecule."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Opacity grid."""
        return self._xsec_grid

    def _load_hdf_file(self, filename: pathlib.Path) -> None:  # noqa: C901
        """Load opacity from HDF5 file.

        Parameters
        ----------
        filename: PathLike
            Path to opacity file.


        """
        import astropy.units as u
        import h5py

        from taurex.data.citation import has_pybtex
        from taurex.util import ensure_string_utf8

        # Load the hdf5
        self.debug(f"Loading opacity from {filename}")

        self._spec_dict = h5py.File(filename, "r")

        self._wavenumber_grid: npt.NDArray[np.float64] = self._spec_dict["bin_edges"][
            ()
        ]

        self._temperature_grid: npt.NDArray[np.float64] = self._spec_dict["t"][()]

        pressure_units = self._spec_dict["p"].attrs["units"]
        try:
            p_conversion = u.Unit(pressure_units).to(u.Pa)
        except UnitConversionError:
            p_conversion = u.Unit(pressure_units, format="cds").to(u.Pa)

        self._pressure_grid: npt.NDArray[np.float64] = (
            self._spec_dict["p"][()] * p_conversion
        )

        if self.in_memory and GlobalCache()["mpi_use_shared"]:
            self._xsec_grid = allocate_as_shared(
                # Dont copy the array until shared memory
                # is allocated. Then copy it to shared memory
                # directly
                self._spec_dict["xsecarr"][()],
                logger=self,
            )
        elif self.in_memory:
            self._xsec_grid = self._spec_dict["xsecarr"][()]
        else:
            self._xsec_grid = self._spec_dict["xsecarr"]

        self._resolution = np.average(np.diff(self._wavenumber_grid))
        self._molecule_name = self._spec_dict["mol_name"][()][()]

        if isinstance(self._molecule_name, np.ndarray):
            self._molecule_name = self._molecule_name[0]

        try:
            self._molecule_name = self._molecule_name.decode()
        except (
            UnicodeDecodeError,
            AttributeError,
        ):
            pass

        self._molecule_name = ensure_string_utf8(self._molecule_name)

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()
        self._molecular_citation = []

        if "DOI" in self._spec_dict:
            doi = self._spec_dict["DOI"][()]
            if isinstance(doi, np.ndarray):
                doi = doi[0]

            molecular_citation = ensure_string_utf8(self._spec_dict["DOI"][()][0])
            self._molecular_citation = [molecular_citation]
        if has_pybtex:
            self.handle_pybtex()

        if self.in_memory:
            self._spec_dict.close()

    def handle_pybtex(self) -> None:
        """Handle citations"""
        from taurex.cache import GlobalCache
        from taurex.data.citation import doi_to_bibtex
        from taurex.util import ensure_string_utf8

        try:
            doi = self._spec_dict["DOI"][()]
            if isinstance(doi, np.ndarray):
                doi = doi[0]

            molecular_citation = ensure_string_utf8(self._spec_dict["DOI"][()][0])
            new_bib = None

            check_xsec = GlobalCache()["xsec_disable_doi"]

            if check_xsec is None:
                check_xsec = False
            else:
                check_xsec = not check_xsec

            if check_xsec:
                new_bib = doi_to_bibtex(molecular_citation)

            self._molecular_citation = [new_bib or molecular_citation]
        except Exception:
            self.warning("DOI could not be determined from hdf5 file.")
            self._molecular_citation = []

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        return self._pressure_grid

    @property
    def resolution(self) -> float:  # noqa: N802
        """Resolution of opacity."""
        return self._resolution

    def citations(self) -> t.List[str]:
        """Return citations for opacity + ExomolOP."""
        from taurex.data.citation import unique_citations_only

        citations = super().citations()
        opacities = []

        for o in self.opacityCitation():
            try:
                from pybtex.database import Entry

                e = Entry.from_string(o, "bibtex")
            except (IndexError, ImportError):
                e = o
            opacities.append(e)

        citations = citations + opacities
        return unique_citations_only(citations)

    def opacityCitation(self) -> t.List[str]:  # noqa: N802
        """Get citations to opacity."""
        return self._molecular_citation

    BIBTEX_ENTRIES = [
        r"""
    @ARTICLE{2021A&A...646A..21C,
        author = {{Chubb}, Katy L. and {Rocchetto}, Marco and {Yurchenko},
        Sergei N. and {Min}, Michiel and {Waldmann}, Ingo and {Barstow}, Joanna K. and
        {Molli{\`e}re}, Paul and {Al-Refaie}, Ahmed F.
        and {Phillips}, Mark W. and {Tennyson},
        Jonathan},
            title = "{The ExoMolOP database: Cross sections and k-tables for
            molecules of interest in high-temperature exoplanet atmospheres}",
        journal = {Astronomy and Astrophysics},
        keywords = {molecular data, opacity, radiative
        transfer, planets and satellites: atmospheres,
        planets and satellites: gaseous planets,
        infrared: planetary systems, Astrophysics - Earth and Planetary Astrophysics,
        Astrophysics - Instrumentation and Methods for
        Astrophysics, Astrophysics - Solar
        and Stellar Astrophysics},
            year = 2021,
            month = feb,
        volume = {646},
            eid = {A21},
            pages = {A21},
            doi = {10.1051/0004-6361/202038350},
    archivePrefix = {arXiv},
        eprint = {2009.00687},
    primaryClass = {astro-ph.EP},
        adsurl = {https://ui.adsabs.harvard.edu/abs/2021A&A...646A..21C},
        adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }

        """
    ]
